"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Statistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class Null(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Unit(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Boolean(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def distribution(self) -> global___Distribution: ...
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            distribution : typing.Optional[global___Distribution] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Id(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Integer(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def distribution(self) -> global___Distribution: ...
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            distribution : typing.Optional[global___Distribution] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Enum(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def distribution(self) -> global___Distribution: ...
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            distribution : typing.Optional[global___Distribution] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Float(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def distribution(self) -> global___Distribution: ...
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            distribution : typing.Optional[global___Distribution] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Text(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        EXAMPLE_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def distribution(self) -> global___Distribution: ...
        example: typing.Text = ...
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            distribution : typing.Optional[global___Distribution] = ...,
            example : typing.Text = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","example",b"example","multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Bytes(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Struct(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class Field(google.protobuf.message.Message):
            """A single field"""
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            STATISTICS_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            @property
            def statistics(self) -> global___Statistics: ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                statistics : typing.Optional[global___Statistics] = ...,
                ) -> None: ...
            def HasField(self, field_name: typing_extensions.Literal["statistics",b"statistics"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing_extensions.Literal["name",b"name","statistics",b"statistics"]) -> None: ...

        FIELDS_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def fields(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Statistics.Struct.Field]:
            """A static struct with named fields"""
            pass
        size: builtins.int = ...
        name: typing.Text = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            fields : typing.Optional[typing.Iterable[global___Statistics.Struct.Field]] = ...,
            size : builtins.int = ...,
            name : typing.Text = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["fields",b"fields","multiplicity",b"multiplicity","name",b"name","size",b"size"]) -> None: ...

    class Union(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class Field(google.protobuf.message.Message):
            """A single field"""
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            STATISTICS_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            @property
            def statistics(self) -> global___Statistics: ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                statistics : typing.Optional[global___Statistics] = ...,
                ) -> None: ...
            def HasField(self, field_name: typing_extensions.Literal["statistics",b"statistics"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing_extensions.Literal["name",b"name","statistics",b"statistics"]) -> None: ...

        FIELDS_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def fields(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Statistics.Union.Field]: ...
        size: builtins.int = ...
        name: typing.Text = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            fields : typing.Optional[typing.Iterable[global___Statistics.Union.Field]] = ...,
            size : builtins.int = ...,
            name : typing.Text = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["fields",b"fields","multiplicity",b"multiplicity","name",b"name","size",b"size"]) -> None: ...

    class Optional(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        STATISTICS_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def statistics(self) -> global___Statistics: ...
        size: builtins.int = ...
        name: typing.Text = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            statistics : typing.Optional[global___Statistics] = ...,
            size : builtins.int = ...,
            name : typing.Text = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["statistics",b"statistics"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["multiplicity",b"multiplicity","name",b"name","size",b"size","statistics",b"statistics"]) -> None: ...

    class List(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        STATISTICS_FIELD_NUMBER: builtins.int
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def statistics(self) -> global___Statistics: ...
        @property
        def distribution(self) -> global___Distribution: ...
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            statistics : typing.Optional[global___Statistics] = ...,
            distribution : typing.Optional[global___Distribution] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution","statistics",b"statistics"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","multiplicity",b"multiplicity","size",b"size","statistics",b"statistics"]) -> None: ...

    class Array(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        STATISTICS_FIELD_NUMBER: builtins.int
        DISTRIBUTIONS_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def statistics(self) -> global___Statistics:
            """A numpy-like n-dimensional array"""
            pass
        @property
        def distributions(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Distribution]: ...
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            statistics : typing.Optional[global___Statistics] = ...,
            distributions : typing.Optional[typing.Iterable[global___Distribution]] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["statistics",b"statistics"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distributions",b"distributions","multiplicity",b"multiplicity","size",b"size","statistics",b"statistics"]) -> None: ...

    class Datetime(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def distribution(self) -> global___Distribution:
            """for a datetime type"""
            pass
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            distribution : typing.Optional[global___Distribution] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Date(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def distribution(self) -> global___Distribution:
            """for a date type"""
            pass
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            distribution : typing.Optional[global___Distribution] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Time(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def distribution(self) -> global___Distribution:
            """for a time type"""
            pass
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            distribution : typing.Optional[global___Distribution] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Duration(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        DISTRIBUTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def distribution(self) -> global___Distribution:
            """for a duration type"""
            pass
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            distribution : typing.Optional[global___Distribution] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["distribution",b"distribution"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["distribution",b"distribution","multiplicity",b"multiplicity","size",b"size"]) -> None: ...

    class Constrained(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        STATISTICS_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        MULTIPLICITY_FIELD_NUMBER: builtins.int
        @property
        def statistics(self) -> global___Statistics: ...
        size: builtins.int = ...
        multiplicity: builtins.float = ...
        def __init__(self,
            *,
            statistics : typing.Optional[global___Statistics] = ...,
            size : builtins.int = ...,
            multiplicity : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["statistics",b"statistics"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["multiplicity",b"multiplicity","size",b"size","statistics",b"statistics"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    NULL_FIELD_NUMBER: builtins.int
    UNIT_FIELD_NUMBER: builtins.int
    BOOLEAN_FIELD_NUMBER: builtins.int
    INTEGER_FIELD_NUMBER: builtins.int
    ENUM_FIELD_NUMBER: builtins.int
    FLOAT_FIELD_NUMBER: builtins.int
    TEXT_FIELD_NUMBER: builtins.int
    BYTES_FIELD_NUMBER: builtins.int
    STRUCT_FIELD_NUMBER: builtins.int
    UNION_FIELD_NUMBER: builtins.int
    OPTIONAL_FIELD_NUMBER: builtins.int
    LIST_FIELD_NUMBER: builtins.int
    ARRAY_FIELD_NUMBER: builtins.int
    DATETIME_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    CONSTRAINED_FIELD_NUMBER: builtins.int
    DATE_FIELD_NUMBER: builtins.int
    TIME_FIELD_NUMBER: builtins.int
    DURATION_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    @property
    def null(self) -> global___Statistics.Null: ...
    @property
    def unit(self) -> global___Statistics.Unit: ...
    @property
    def boolean(self) -> global___Statistics.Boolean: ...
    @property
    def integer(self) -> global___Statistics.Integer: ...
    @property
    def enum(self) -> global___Statistics.Enum: ...
    @property
    def float(self) -> global___Statistics.Float: ...
    @property
    def text(self) -> global___Statistics.Text: ...
    @property
    def bytes(self) -> global___Statistics.Bytes: ...
    @property
    def struct(self) -> global___Statistics.Struct: ...
    @property
    def union(self) -> global___Statistics.Union: ...
    @property
    def optional(self) -> global___Statistics.Optional: ...
    @property
    def list(self) -> global___Statistics.List: ...
    @property
    def array(self) -> global___Statistics.Array: ...
    @property
    def datetime(self) -> global___Statistics.Datetime: ...
    @property
    def id(self) -> global___Statistics.Id: ...
    @property
    def constrained(self) -> global___Statistics.Constrained: ...
    @property
    def date(self) -> global___Statistics.Date: ...
    @property
    def time(self) -> global___Statistics.Time: ...
    @property
    def duration(self) -> global___Statistics.Duration: ...
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """Other properties"""
        pass
    def __init__(self,
        *,
        name : typing.Text = ...,
        null : typing.Optional[global___Statistics.Null] = ...,
        unit : typing.Optional[global___Statistics.Unit] = ...,
        boolean : typing.Optional[global___Statistics.Boolean] = ...,
        integer : typing.Optional[global___Statistics.Integer] = ...,
        enum : typing.Optional[global___Statistics.Enum] = ...,
        float : typing.Optional[global___Statistics.Float] = ...,
        text : typing.Optional[global___Statistics.Text] = ...,
        bytes : typing.Optional[global___Statistics.Bytes] = ...,
        struct : typing.Optional[global___Statistics.Struct] = ...,
        union : typing.Optional[global___Statistics.Union] = ...,
        optional : typing.Optional[global___Statistics.Optional] = ...,
        list : typing.Optional[global___Statistics.List] = ...,
        array : typing.Optional[global___Statistics.Array] = ...,
        datetime : typing.Optional[global___Statistics.Datetime] = ...,
        id : typing.Optional[global___Statistics.Id] = ...,
        constrained : typing.Optional[global___Statistics.Constrained] = ...,
        date : typing.Optional[global___Statistics.Date] = ...,
        time : typing.Optional[global___Statistics.Time] = ...,
        duration : typing.Optional[global___Statistics.Duration] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["array",b"array","boolean",b"boolean","bytes",b"bytes","constrained",b"constrained","date",b"date","datetime",b"datetime","duration",b"duration","enum",b"enum","float",b"float","id",b"id","integer",b"integer","list",b"list","null",b"null","optional",b"optional","statistics",b"statistics","struct",b"struct","text",b"text","time",b"time","union",b"union","unit",b"unit"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["array",b"array","boolean",b"boolean","bytes",b"bytes","constrained",b"constrained","date",b"date","datetime",b"datetime","duration",b"duration","enum",b"enum","float",b"float","id",b"id","integer",b"integer","list",b"list","name",b"name","null",b"null","optional",b"optional","properties",b"properties","statistics",b"statistics","struct",b"struct","text",b"text","time",b"time","union",b"union","unit",b"unit"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["statistics",b"statistics"]) -> typing.Optional[typing_extensions.Literal["null","unit","boolean","integer","enum","float","text","bytes","struct","union","optional","list","array","datetime","id","constrained","date","time","duration"]]: ...
global___Statistics = Statistics

class Distribution(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class Double(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class Point(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            VALUE_FIELD_NUMBER: builtins.int
            PROBABILITY_FIELD_NUMBER: builtins.int
            value: builtins.float = ...
            probability: builtins.float = ...
            def __init__(self,
                *,
                value : builtins.float = ...,
                probability : builtins.float = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["probability",b"probability","value",b"value"]) -> None: ...

        POINTS_FIELD_NUMBER: builtins.int
        MIN_FIELD_NUMBER: builtins.int
        MAX_FIELD_NUMBER: builtins.int
        @property
        def points(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Distribution.Double.Point]: ...
        min: builtins.float = ...
        max: builtins.float = ...
        def __init__(self,
            *,
            points : typing.Optional[typing.Iterable[global___Distribution.Double.Point]] = ...,
            min : builtins.float = ...,
            max : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["max",b"max","min",b"min","points",b"points"]) -> None: ...

    class Integer(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class Point(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            VALUE_FIELD_NUMBER: builtins.int
            PROBABILITY_FIELD_NUMBER: builtins.int
            value: builtins.int = ...
            probability: builtins.float = ...
            def __init__(self,
                *,
                value : builtins.int = ...,
                probability : builtins.float = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["probability",b"probability","value",b"value"]) -> None: ...

        POINTS_FIELD_NUMBER: builtins.int
        MIN_FIELD_NUMBER: builtins.int
        MAX_FIELD_NUMBER: builtins.int
        @property
        def points(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Distribution.Integer.Point]: ...
        min: builtins.int = ...
        max: builtins.int = ...
        def __init__(self,
            *,
            points : typing.Optional[typing.Iterable[global___Distribution.Integer.Point]] = ...,
            min : builtins.int = ...,
            max : builtins.int = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["max",b"max","min",b"min","points",b"points"]) -> None: ...

    class Boolean(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class Point(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            PROBABILITY_FIELD_NUMBER: builtins.int
            name: builtins.bool = ...
            value: builtins.int = ...
            probability: builtins.float = ...
            def __init__(self,
                *,
                name : builtins.bool = ...,
                value : builtins.int = ...,
                probability : builtins.float = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["name",b"name","probability",b"probability","value",b"value"]) -> None: ...

        POINTS_FIELD_NUMBER: builtins.int
        @property
        def points(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Distribution.Boolean.Point]: ...
        def __init__(self,
            *,
            points : typing.Optional[typing.Iterable[global___Distribution.Boolean.Point]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["points",b"points"]) -> None: ...

    class Enum(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class Point(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            PROBABILITY_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            value: builtins.float = ...
            probability: builtins.float = ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                value : builtins.float = ...,
                probability : builtins.float = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["name",b"name","probability",b"probability","value",b"value"]) -> None: ...

        POINTS_FIELD_NUMBER: builtins.int
        @property
        def points(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Distribution.Enum.Point]: ...
        def __init__(self,
            *,
            points : typing.Optional[typing.Iterable[global___Distribution.Enum.Point]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["points",b"points"]) -> None: ...

    DOUBLE_FIELD_NUMBER: builtins.int
    INTEGER_FIELD_NUMBER: builtins.int
    BOOLEAN_FIELD_NUMBER: builtins.int
    ENUM_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    @property
    def double(self) -> global___Distribution.Double: ...
    @property
    def integer(self) -> global___Distribution.Integer: ...
    @property
    def boolean(self) -> global___Distribution.Boolean: ...
    @property
    def enum(self) -> global___Distribution.Enum: ...
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """Needed for compatibility with typing"""
        pass
    def __init__(self,
        *,
        double : typing.Optional[global___Distribution.Double] = ...,
        integer : typing.Optional[global___Distribution.Integer] = ...,
        boolean : typing.Optional[global___Distribution.Boolean] = ...,
        enum : typing.Optional[global___Distribution.Enum] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["boolean",b"boolean","distribution",b"distribution","double",b"double","enum",b"enum","integer",b"integer"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["boolean",b"boolean","distribution",b"distribution","double",b"double","enum",b"enum","integer",b"integer","properties",b"properties"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["distribution",b"distribution"]) -> typing.Optional[typing_extensions.Literal["double","integer","boolean","enum"]]: ...
global___Distribution = Distribution



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/proto_container.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.sarus_data_spec/protobuf/proto_container.proto\x12\x0fsarus_data_spec\x1a\x19google/protobuf/any.proto\"\xae\x01\n\x0eProtoContainer\x12$\n\x06protos\x18\x01 \x03(\x0b\x32\x14.google.protobuf.Any\x12\x43\n\nproperties\x18\x02 \x03(\x0b\x32/.sarus_data_spec.ProtoContainer.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.proto_container_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PROTOCONTAINER_PROPERTIESENTRY._options = None
  _PROTOCONTAINER_PROPERTIESENTRY._serialized_options = b'8\001'
  _PROTOCONTAINER._serialized_start=95
  _PROTOCONTAINER._serialized_end=269
  _PROTOCONTAINER_PROPERTIESENTRY._serialized_start=220
  _PROTOCONTAINER_PROPERTIESENTRY._serialized_end=269
# @@protoc_insertion_point(module_scope)



import hashlib
import logging
import typing as t
import warnings

import pyarrow as pa

from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.ops.base import (
    DatasetImplementation,
    DatasetStaticChecker,
    DataspecStaticChecker,
    ScalarImplementation,
    _ensure_batch_correct,
)
from sarus_data_spec.typing import Path

try:
    from sarus_data_spec.manager.ops.sql_utils.table_mapping import (
        name_encoder,
        table_mapping,
    )
except ModuleNotFoundError:
    warnings.warn("table_mapping not found. Cannot send sql queries.")
try:
    from sarus_data_spec.manager.ops.sql_utils.queries import (
        rename_and_compose_queries,
    )
except ModuleNotFoundError:
    warnings.warn("Queries composition not available.")
from sarus_data_spec.scalar import Scalar
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)


class StandardDatasetStaticChecker(DatasetStaticChecker):
    def parent(self, kind: str = "dataset") -> t.Union[st.Dataset, st.Scalar]:
        return parent(self.dataset, kind=kind)

    async def parent_schema(self) -> st.Schema:
        parent = self.parent(kind="dataset")
        assert isinstance(parent, Dataset)
        return await parent.manager().async_schema(parent)

    async def parent_marginals(self) -> st.Marginals:
        parent = self.parent(kind="dataset")
        assert isinstance(parent, Dataset)
        return await parent.manager().async_marginals(parent)

    def pup_token(self, public_context: t.Collection[str]) -> t.Optional[str]:
        """By default we implement that the transform inherits the PUP status
        but changes the PUP token."""
        parent_token = self.parent().pup_token()
        if parent_token is None:
            return None

        transform = self.dataset.transform()
        h = hashlib.md5(usedforsecurity=False)
        h.update(parent_token.encode("ascii"))
        h.update(transform.protobuf().SerializeToString())

        return h.hexdigest()

    def rewritten_pup_token(
        self, public_context: t.Collection[str]
    ) -> t.Optional[str]:
        """By default we implement that the transform inherits the PUP status
        but changes the PUP token."""
        parent_token = self.parent().rewritten_pup_token()
        if parent_token is None:
            return None

        transform = self.dataset.transform()
        h = hashlib.md5(usedforsecurity=False)
        h.update(parent_token.encode("ascii"))
        h.update(transform.protobuf().SerializeToString())

        return h.hexdigest()


class StandardDatasetImplementation(DatasetImplementation):
    """Object that executes first routing among ops between
    transformed/source and processor
    """

    def parents(self) -> t.List[t.Union[st.DataSpec, st.Transform]]:
        return parents(self.dataset)

    def parent(self, kind: str = "dataset") -> t.Union[st.Dataset, st.Scalar]:
        return parent(self.dataset, kind=kind)

    async def parent_to_arrow(
        self, batch_size: int = 10000
    ) -> t.AsyncIterator[pa.RecordBatch]:
        parent = self.parent(kind="dataset")
        assert isinstance(parent, Dataset)
        parent_iterator = await parent.manager().async_to_arrow(
            parent, batch_size=batch_size
        )
        return await self.decoupled_async_iter(parent_iterator)

    async def parent_schema(self) -> st.Schema:
        parent = self.parent(kind="dataset")
        assert isinstance(parent, Dataset)
        return await parent.manager().async_schema(parent)

    async def parent_value(self) -> t.Any:
        parent = self.parent(kind="scalar")
        assert isinstance(parent, Scalar)
        return await parent.manager().async_value(parent)

    async def parent_size(self) -> st.Size:
        parent = self.parent(kind="dataset")
        assert isinstance(parent, Dataset)
        return await parent.manager().async_size(parent)

    async def parent_multiplicity(self) -> st.Multiplicity:
        parent = self.parent(kind="dataset")
        assert isinstance(parent, Dataset)
        return await parent.manager().async_multiplicity(parent)

    async def parent_bounds(self) -> st.Bounds:
        parent = self.parent(kind="dataset")
        assert isinstance(parent, Dataset)
        return await parent.manager().async_bounds(parent)

    async def parent_marginals(self) -> st.Marginals:
        parent = self.parent(kind="dataset")
        assert isinstance(parent, Dataset)
        return await parent.manager().async_marginals(parent)

    async def ensure_batch_correct(
        self,
        async_iterator: t.AsyncIterator[pa.RecordBatch],
        func_to_apply: t.Callable,
        batch_size: int,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        """Method that executes func_to_apply on each batch
        of the async_iterator but rather than directly returning
        the result, it accumulates them and returns them progressively
        so that each new batch has batch_size."""

        return _ensure_batch_correct(async_iterator, func_to_apply, batch_size)

    async def sql_implementation(
        self,
    ) -> t.Optional[t.Dict[t.Tuple[str, ...], str]]:
        """Returns a dict of queries equivalent to the current transform.
        If the the transform does not change the schema, then return None"""
        raise NotImplementedError(
            "No SQL implementation for dataset issued from"
            f" {self.dataset.transform().spec()} transform."
        )

    async def sql(
        self,
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
        result_type: t.Optional[st.Type] = None,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        """It rewrites and/or composes the query and sends it to the parent."""
        queries_transform = await self.sql_implementation()
        current_schema = await self.dataset.manager().async_schema(
            self.dataset
        )
        parent_schema = await self.parent_schema()
        if (
            queries_transform is None
            and current_schema.name() == parent_schema.name()
        ):
            parent_query = query
        else:
            table_map: dict[Path, tuple[str, ...]] = {}
            if queries_transform is not None:
                table_map = t.cast(
                    dict[Path, tuple[str, ...]],
                    table_mapping(
                        tables=current_schema.tables(),
                        sarus_schema_name=current_schema.name(),
                        encoded_name_length=10,
                        encoder_prefix_name=self.dataset.uuid(),
                    ),
                )
            updated_queries_transform = (
                {
                    name_encoder(
                        names=(self.dataset.uuid(), *tab_name),
                        length=10,
                    ): query_str
                    for tab_name, query_str in queries_transform.items()
                }
                if queries_transform is not None
                else None
            )

            parent_query = rename_and_compose_queries(
                query_or_dict=query,
                curr_path=[],
                queries_transform=updated_queries_transform,
                table_map=table_map,
            )

        parent_ds = t.cast(st.Dataset, self.parent(kind="dataset"))
        logger.info(
            f"query {parent_query} sent to the "
            f"parent dataset {parent_ds.uuid()}"
        )
        return await parent_ds.manager().async_sql(
            dataset=parent_ds,
            query=parent_query,
            dialect=dialect,
            batch_size=batch_size,
            result_type=result_type,
        )


class StandardScalarStaticChecker(DataspecStaticChecker): ...


class StandardScalarImplementation(ScalarImplementation):
    def parent(self, kind: str = "dataset") -> st.DataSpec:
        return parent(self.scalar, kind=kind)

    def parents(self) -> t.List[t.Union[st.DataSpec, st.Transform]]:
        return parents(self.scalar)

    async def parent_to_arrow(
        self, batch_size: int = 10000
    ) -> t.AsyncIterator[pa.RecordBatch]:
        parent = self.parent(kind="dataset")
        assert isinstance(parent, Dataset)
        parent_iterator = await parent.manager().async_to_arrow(
            parent, batch_size=batch_size
        )
        return await self.decoupled_async_iter(parent_iterator)

    async def parent_schema(self) -> st.Schema:
        parent = self.parent(kind="dataset")
        assert isinstance(parent, Dataset)
        return await parent.manager().async_schema(parent)

    async def parent_value(self) -> t.Any:
        parent = self.parent(kind="scalar")
        assert isinstance(parent, Scalar)
        return await parent.manager().async_value(parent)


def parent(dataspec: st.DataSpec, kind: str) -> t.Union[st.Dataset, st.Scalar]:
    pars = parents(dataspec)
    if kind == "dataset":
        parent: t.Union[t.List[Scalar], t.List[Dataset]] = [
            element for element in pars if isinstance(element, Dataset)
        ]
    else:
        parent = [element for element in pars if isinstance(element, Scalar)]
    assert len(parent) == 1
    return parent[0]


def parents(
    dataspec: st.DataSpec,
) -> t.List[t.Union[st.DataSpec, st.Transform]]:
    parents_args, parents_kwargs = dataspec.parents()
    parents_args.extend(parents_kwargs.values())
    return parents_args



import asyncio
import asyncio.events as events
import typing as t

import pyarrow as pa

T = t.TypeVar("T")


def sync(coro: t.Coroutine) -> t.Any:
    """This runs an async function synchronously,
    even within an already runnning event loop,
    despite the apparent impossibility
    to nest loops within the same thread"""
    reset_at_completion = True
    try:
        curr_loop = asyncio.get_running_loop()
    except RuntimeError:
        reset_at_completion = False
    new_loop = asyncio.new_event_loop()
    new_loop._check_running = lambda: None  # type:ignore
    # method monkey patched since it raises an error
    # if another loop exists
    result = new_loop.run_until_complete(coro)
    if reset_at_completion:
        # this method allows to set the curr loop
        # as running,standard method asyncio.set_event_loop
        # does not work
        events._set_running_loop(curr_loop)
    new_loop.close()
    return result


def sync_iterator_from_async_iterator(
    async_iterator: t.AsyncIterator[T],
) -> t.Iterator[T]:
    """This methods returns an iterator from an async iterator.
    The generator method allows to execute one step at a time the
    anext method of the async generator.
    """
    # The loop used to generate batches
    new_loop = asyncio.new_event_loop()
    new_loop._check_running = lambda: None  # type:ignore
    # method monkey patched since it raises an error
    # if another loop exists

    def generator() -> t.Iterator[T]:
        """This method creates an iterator. At each generator
        step, the current loop is taken and suspended and the
        anext method of the async generator is executed in
        another loop
        """
        keep_on = True
        while keep_on:
            reset_at_completion = True
            try:
                curr_loop = asyncio.get_running_loop()
            except RuntimeError:
                reset_at_completion = False
            try:
                batch = new_loop.run_until_complete(async_iterator.__anext__())

            except StopAsyncIteration:
                keep_on = False
                if reset_at_completion:
                    events._set_running_loop(curr_loop)
            except Exception as exception:
                if reset_at_completion:
                    events._set_running_loop(curr_loop)
                new_loop.close()
                raise exception
            else:
                if reset_at_completion:
                    # asyncio.set_event_loop(curr_loop) not working
                    events._set_running_loop(curr_loop)
                yield batch
        new_loop.close()

    return generator()


def sync_iterator(
    async_iterator_coro: t.Coroutine,
) -> t.Iterator[t.Any]:
    """This methods returns an iterator from an async iterator coroutine.
    It first executes the coroutine to obtain an async generator.
    Then, generator method allows to execute one step at a time the
    anext method of the async generator.
    """
    async_gen = sync(async_iterator_coro)
    return sync_iterator_from_async_iterator(async_gen)


async def async_iter(data_list: t.Collection[T]) -> t.AsyncIterator[T]:
    """Convert a collection into an AsyncIterator."""
    for data in data_list:
        yield data


async def decoupled_async_iter(
    source: t.AsyncIterator[T], buffer_size: int = 100
) -> t.AsyncIterator[T]:
    """Create a consumer/producer pattern using an asyncio.Queue."""
    queue: asyncio.Queue = asyncio.Queue(maxsize=buffer_size)

    async def producer() -> None:
        async for x in source:
            await queue.put(x)
        await queue.put(None)  # producer finished

    # Launch the iteration of source iterator
    loop = asyncio.get_running_loop()
    loop.create_task(producer())

    while True:
        x = await queue.get()
        if x is None:
            queue.task_done()
            break
        queue.task_done()
        yield x


def to_recordbatch(
    output: t.Mapping[str, pa.Table],
) -> t.AsyncIterator[pa.RecordBatch]:
    """This function is inspired from arrow_recordbatch in
    sarus_data_spec/manager/ops/source/sql/arrow.py
    In order to create a recordbatch from different pyarrow tables, a struct is
    created where the corresponding other missing tables
    have been added as None"""
    struct_arrays = {
        table_path: pa.StructArray.from_arrays(
            arrays=[el.combine_chunks() for el in table.flatten()],
            names=table.column_names,
        )
        for table_path, table in output.items()
    }
    batches = []
    names = list(struct_arrays.keys())

    for current_table, arr_array in struct_arrays.items():
        # iterate over tables
        arrays_list = [
            pa.array(
                [None] * len(arr_array),
                type=struct.type,
            )
            if current_table != table
            else arr_array
            for table, struct in struct_arrays.items()
        ]
        fields = [
            pa.field(name=name, type=arr.type)
            for name, arr in zip(names, arrays_list)
        ]
        fields.append(
            pa.field(
                name="field_selected", type=pa.large_string(), nullable=False
            )
        )
        arrays_list.append(
            pa.array([current_table] * len(arr_array), type=pa.large_string())
        )
        batches.append(
            pa.RecordBatch.from_arrays(
                arrays_list, schema=pa.schema(fields=fields)
            )
        )
    return async_iter(batches)


def standardize_sql_query(
    parent_schema_name: str,
    query: t.Union[str, t.Mapping[t.Union[str, t.Tuple[str, ...]], str]],
) -> t.Mapping[str, str]:
    """It transforms any query variant (string, dict etc) in
    t.Mapping[str, str]. Moreover it applies the rules to preserve or not
    the parent schema in the map key which will be used as parecordbach name.

    If query is:
    {'a.b.c': sqlquerystr, 'a.d.e': sqlquerystr} -> override parent schema name
    {'a.b': sqlquerystr, 'a.c': sqlquerystr} override parent schema name
    {'b': sqlquerystr, 'c': sqlquerystr} preserve parent schema name
    {'a.b.c': sqlquerystr, 'd.e': sqlquerystr} preserve parent schema name
    {'a': 'q'} -> override parent schema name
    {'a.b': 'q'} -> override parent schema name
    sqlquerystr  -> preserve parent schema name
    """

    if isinstance(query, str):
        # we preserve the schema name of the parent
        queries = {parent_schema_name: query}
    else:
        if len(query) == 1:
            # we override the schema name
            queries = {
                key if isinstance(key, str) else ".".join(key): value
                for key, value in query.items()
            }
        else:
            # we override if prefix is the same otherwise we preserve
            prefixes = [
                key.split(".")[0] if isinstance(key, str) else key[0]
                for key in list(query.keys())
            ]
            if len(prefixes) == len(set(prefixes)):
                # all prefixes are the same, we preserve parents schema
                queries = {
                    f"{parent_schema_name}.{key}"
                    if isinstance(key, str)
                    else ".".join([parent_schema_name, *key]): value
                    for key, value in query.items()
                }
            else:
                # we override parents schema
                queries = {
                    key if isinstance(key, str) else ".".join(key): value
                    for key, value in query.items()
                }
    return queries



from collections import defaultdict
import logging
import typing as t

try:
    from sarus_sql import ast_utils, rename_tables, translator
except ModuleNotFoundError:
    logger = logging.getLogger(__name__)
    logger.info("sarus_sql not available.")

from sarus_data_spec.typing import Path
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st


def rename_and_compose_queries(
    query_or_dict: t.Union[str, t.Dict[str, t.Any]],
    curr_path: t.List[str],
    queries_transform: t.Optional[t.Dict[str, str]],
    table_map: t.Dict[Path, t.Tuple[str, ...]],
) -> t.Union[str, t.Dict[str, t.Any]]:
    """Composition is done by first updating the table names
    in the leaves of queries_or_dict and then composing with
    all the queries in queries_transform
    """
    if isinstance(query_or_dict, str):
        # type ignore because issue of typing in sarus_sql
        updated_query: str = rename_tables.rename_tables(
            query_or_dict, table_map
        )  # noqa : E501
        if queries_transform is not None:
            updated_query = ast_utils.compose_query(
                queries_transform, updated_query
            )
        return updated_query
    else:
        new_queries: dict[str, t.Any] = {}
        for name, sub_queries_or_dict in query_or_dict.items():
            new_queries[name] = rename_and_compose_queries(
                query_or_dict=sub_queries_or_dict,
                curr_path=[*curr_path, name],
                queries_transform=queries_transform,
                table_map=table_map,
            )
        return new_queries


def flatten_queries_dict(
    queries: t.Dict[str, t.Any],
) -> t.Dict[t.Tuple[str, ...], str]:
    """Transform nested dict in linear dict where each
    key is the tuple of the nesting path"""

    final_dict: t.Dict[t.Tuple[str, ...], str] = {}

    def update_dict(
        curr_path: t.List[str],
        dict_to_update: t.Dict[t.Tuple[str, ...], t.Any],
        query_or_dict: t.Union[t.Dict[str, t.Any], t.Any],
    ) -> None:
        if isinstance(query_or_dict, dict):
            for name, sub_query in query_or_dict.items():
                update_dict(
                    curr_path=[*curr_path, name],
                    dict_to_update=dict_to_update,
                    query_or_dict=sub_query,
                )
        else:
            dict_to_update[tuple(curr_path)] = t.cast(str, query_or_dict)
        return

    for name, query_or_dict in queries.items():
        update_dict(
            query_or_dict=query_or_dict,
            curr_path=[name],
            dict_to_update=final_dict,
        )

    return final_dict


def rename_and_translate_query(
    old_query: str,
    dialect: st.SQLDialect,
    destination_dialect: st.SQLDialect,
    table_mapping: t.Dict[st.Path, t.Tuple[str]],
    extended_table_mapping: t.Optional[t.List[str]],
) -> str:
    """Converts to postgres, parses query and then
    reconverts if needed"""

    # Translate to postgres
    new_query = str(
        translator.translate_to_postgres(
            ast_utils.parse_query(old_query),
            dialect,
        )
    )
    # Rename tables
    new_query = rename_tables.rename_tables(
        query_str=new_query,
        table_mapping=t.cast(
            t.Dict[st.Path, t.Tuple[str, ...]], table_mapping
        ),
        extended_table_mapping=extended_table_mapping,
    )
    if destination_dialect != st.SQLDialect.POSTGRES:
        new_query = str(
            translator.translate_to_dialect(
                ast_utils.parse_query(new_query),
                destination_dialect,
            )
        )
    return new_query


def nest_queries(
    queries: t.Dict[t.Tuple[str, ...], str],
) -> t.Dict[str, t.Any]:
    """It transform the dict of queries according to the tuple keys:
    if queries = {
            ('a','b'):'q',
            ('a','c'):'q'
    }
    the results woulf be: {a: {b: 'q', c: 'q'}

    if queries = {
            ('a','b'):'q',
            ('e','c'):'q'
    }
    the results woulf be: {a: {b: 'q'}, e: {c: 'q'}}
    """
    intermediate: t.Dict[str, t.Dict[t.Tuple[str, ...], t.Any]] = defaultdict(
        dict
    )
    final: t.Dict[str, t.Any] = {}
    for query_path, query in queries.items():
        if len(query_path) == 0:
            final[""] = query
        elif len(query_path) == 1:
            final[query_path[0]] = query
        else:
            intermediate[query_path[0]][query_path[1:]] = query

    for name, subdict in intermediate.items():
        final[name] = nest_queries(subdict)

    return final


def nested_dict_of_types(
    types: t.Dict[t.Tuple[str, ...], st.Type],
) -> t.Dict[str, t.Any]:
    """Similar to nest_queries but values are sarus types instead of strings"""
    intermediate: t.Dict[str, t.Dict[t.Tuple[str, ...], t.Any]] = defaultdict(
        dict
    )
    final: t.Dict[str, t.Any] = {}
    for type_path, type in types.items():
        if len(type_path) == 1:
            final[type_path[0]] = type
        else:
            intermediate[type_path[0]][type_path[1:]] = type

    for name, subdict in intermediate.items():
        final[name] = nested_dict_of_types(subdict)

    return final


def nested_unions_from_nested_dict_of_types(
    nested_types: t.Dict[str, t.Any],
) -> t.Dict[str, st.Type]:
    """create unions out of nested_types"""
    fields: t.Dict[str, st.Type] = {}
    for path_string, type_or_dict in nested_types.items():
        if isinstance(type_or_dict, dict):
            fields[path_string] = sdt.Union(
                nested_unions_from_nested_dict_of_types(type_or_dict)
            )
        else:
            fields[path_string] = type_or_dict
    return fields



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/transform.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import type_pb2 as sarus__data__spec_dot_protobuf_dot_type__pb2
from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2
from sarus_data_spec.protobuf import scalar_pb2 as sarus__data__spec_dot_protobuf_dot_scalar__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(sarus_data_spec/protobuf/transform.proto\x1a#sarus_data_spec/protobuf/type.proto\x1a#sarus_data_spec/protobuf/path.proto\x1a%sarus_data_spec/protobuf/scalar.proto\"\xbb+\n\tTransform\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12\x1d\n\x04spec\x18\x04 \x01(\x0b\x32\x0f.Transform.Spec\x12.\n\nproperties\x18\x05 \x03(\x0b\x32\x1a.Transform.PropertiesEntry\x12\x12\n\ninversible\x18\x06 \x01(\x08\x12\x19\n\x11schema_preserving\x18\x07 \x01(\x08\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xea\x0e\n\x04Spec\x12\'\n\x08identity\x18\x01 \x01(\x0b\x32\x13.Transform.IdentityH\x00\x12\'\n\x08variable\x18\x02 \x01(\x0b\x32\x13.Transform.VariableH\x00\x12\'\n\x08\x63omposed\x18\x03 \x01(\x0b\x32\x13.Transform.ComposedH\x00\x12%\n\x07project\x18\x04 \x01(\x0b\x32\x12.Transform.ProjectH\x00\x12#\n\x06\x66ilter\x18\x05 \x01(\x0b\x32\x11.Transform.FilterH\x00\x12%\n\x07shuffle\x18\x06 \x01(\x0b\x32\x12.Transform.ShuffleH\x00\x12\x1f\n\x04join\x18\x07 \x01(\x0b\x32\x0f.Transform.JoinH\x00\x12\x1f\n\x04\x63\x61st\x18\x08 \x01(\x0b\x32\x0f.Transform.CastH\x00\x12#\n\x06sample\x18\t \x01(\x0b\x32\x11.Transform.SampleH\x00\x12\x30\n\ruser_settings\x18\n \x01(\x0b\x32\x17.Transform.UserSettingsH\x00\x12?\n\x15privacy_unit_tracking\x18\x0b \x01(\x0b\x32\x1e.Transform.PrivacyUnitTrackingH\x00\x12\'\n\x08\x65xternal\x18\x0c \x01(\x0b\x32\x13.Transform.ExternalH\x00\x12)\n\tsynthetic\x18\r \x01(\x0b\x32\x14.Transform.SyntheticH\x00\x12)\n\ttranscode\x18\x0e \x01(\x0b\x32\x14.Transform.TranscodeH\x00\x12\x38\n\x11inverse_transcode\x18\x0f \x01(\x0b\x32\x1b.Transform.InverseTranscodeH\x00\x12&\n\x08get_item\x18\x10 \x01(\x0b\x32\x12.Transform.GetItemH\x00\x12J\n\x1bprivacy_unit_tracking_paths\x18\x11 \x01(\x0b\x32#.Transform.PrivacyUnitTrackingPathsH\x00\x12\x43\n\x17\x61utomatic_user_settings\x18\x12 \x01(\x0b\x32 .Transform.AutomaticUserSettingsH\x00\x12.\n\x0cpublic_paths\x18\x13 \x01(\x0b\x32\x16.Transform.PublicPathsH\x00\x12\x30\n\rassign_budget\x18\x14 \x01(\x0b\x32\x17.Transform.AssignBudgetH\x00\x12\x36\n\x10\x61utomatic_budget\x18\x15 \x01(\x0b\x32\x1a.Transform.AutomaticBudgetH\x00\x12\x37\n\x10\x61ttribute_budget\x18\x16 \x01(\x0b\x32\x1b.Transform.AttributesBudgetH\x00\x12(\n\tsd_budget\x18\x17 \x01(\x0b\x32\x13.Transform.SDBudgetH\x00\x12,\n\x0b\x64\x65rive_seed\x18\x18 \x01(\x0b\x32\x15.Transform.DeriveSeedH\x00\x12+\n\x0bgroup_by_pe\x18\x19 \x01(\x0b\x32\x14.Transform.GroupByPEH\x00\x12\x34\n\x0fsampling_ratios\x18\x1a \x01(\x0b\x32\x19.Transform.SamplingRatiosH\x00\x12*\n\nselect_sql\x18\x1b \x01(\x0b\x32\x14.Transform.SelectSqlH\x00\x12%\n\x07\x65xtract\x18\x1c \x01(\x0b\x32\x12.Transform.ExtractH\x00\x12\x38\n\x11relationship_spec\x18\x1d \x01(\x0b\x32\x1b.Transform.RelationshipSpecH\x00\x12@\n\x15\x64ifferentiated_sample\x18\x1e \x01(\x0b\x32\x1f.Transform.DifferentiatedSampleH\x00\x12;\n\x13validated_user_type\x18\x1f \x01(\x0b\x32\x1c.Transform.ValidatedUserTypeH\x00\x12/\n\rdp_select_sql\x18  \x01(\x0b\x32\x16.Transform.DPSelectSqlH\x00\x12\x36\n\x10\x65rror_estimation\x18! \x01(\x0b\x32\x1a.Transform.ErrorEstimationH\x00\x12(\n\tfit_model\x18\" \x01(\x0b\x32\x13.Transform.FitModelH\x00\x12-\n\x0c\x66it_model_dp\x18# \x01(\x0b\x32\x15.Transform.FitModelDPH\x00\x12;\n\x13generate_from_model\x18$ \x01(\x0b\x32\x1c.Transform.GenerateFromModelH\x00\x12/\n\rto_small_data\x18% \x01(\x0b\x32\x16.Transform.ToSmallDataH\x00\x12&\n\x08push_sql\x18& \x01(\x0b\x32\x12.Transform.PushSQLH\x00\x12.\n\x0cselect_table\x18\' \x01(\x0b\x32\x16.Transform.SelectTableH\x00\x42\x06\n\x04spec\x1a\xf6\x07\n\x08\x45xternal\x12\x11\n\targuments\x18\x01 \x01(\x0c\x12\x17\n\x0fnamed_arguments\x18\x02 \x01(\x0c\x12\x37\n\rop_identifier\x18\x03 \x01(\x0b\x32 .Transform.External.OpIdentifier\x1a\xd7\x04\n\x0cOpIdentifier\x12&\n\x03std\x18\x01 \x01(\x0b\x32\x17.Transform.External.StdH\x00\x12,\n\x06pandas\x18\x02 \x01(\x0b\x32\x1a.Transform.External.PandasH\x00\x12*\n\x05numpy\x18\x03 \x01(\x0b\x32\x19.Transform.External.NumpyH\x00\x12\x34\n\ntensorflow\x18\x04 \x01(\x0b\x32\x1e.Transform.External.TensorflowH\x00\x12.\n\x07sklearn\x18\x05 \x01(\x0b\x32\x1b.Transform.External.SklearnH\x00\x12?\n\x10pandas_profiling\x18\x06 \x01(\x0b\x32#.Transform.External.PandasProfilingH\x00\x12.\n\x07xgboost\x18\x07 \x01(\x0b\x32\x1b.Transform.External.XGBoostH\x00\x12*\n\x05skopt\x18\x08 \x01(\x0b\x32\x19.Transform.External.SkoptH\x00\x12\x30\n\x08imblearn\x18\t \x01(\x0b\x32\x1c.Transform.External.ImblearnH\x00\x12(\n\x04shap\x18\n \x01(\x0b\x32\x18.Transform.External.ShapH\x00\x12*\n\x05scipy\x18\x0b \x01(\x0b\x32\x19.Transform.External.ScipyH\x00\x12\x34\n\noptbinning\x18\x0c \x01(\x0b\x32\x1e.Transform.External.OptBinningH\x00\x42\x04\n\x02op\x1a\x13\n\x03Std\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x16\n\x06Pandas\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Numpy\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1a\n\nTensorflow\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x17\n\x07Sklearn\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1f\n\x0fPandasProfiling\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x17\n\x07XGBoost\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Skopt\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x18\n\x08Imblearn\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x14\n\x04Shap\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Scipy\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1a\n\nOptBinning\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\n\n\x08Identity\x1a*\n\x08Variable\x12\x10\n\x08position\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x1a\xa9\x01\n\x08\x43omposed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12@\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\'.Transform.Composed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x34\n\x07Project\x12)\n\nprojection\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a/\n\x06\x46ilter\x12%\n\x06\x66ilter\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a\t\n\x07Shuffle\x1a)\n\x04Join\x12!\n\x02on\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a+\n\x04\x43\x61st\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a\x61\n\x06Sample\x12\x12\n\x08\x66raction\x18\x01 \x01(\x01H\x00\x12\x0e\n\x04size\x18\x02 \x01(\x03H\x00\x12%\n\x04seed\x18\x03 \x01(\x0b\x32\x17.sarus_data_spec.ScalarB\x0c\n\nproportion\x1av\n\x0fSchemaInference\x12:\n\x0b\x63\x61st_policy\x18\x01 \x01(\x0e\x32%.Transform.SchemaInference.CastPolicy\"\'\n\nCastPolicy\x12\x08\n\x04NONE\x10\x00\x12\x0f\n\x0bMOST_LIKELY\x10\x01\x1a\x16\n\x07GroupBy\x12\x0b\n\x03key\x18\x01 \x01(\t\x1a[\n\x0bToSmallData\x12\x0c\n\x04size\x18\x01 \x01(\x05\x12\x17\n\x0frandom_sampling\x18\x02 \x01(\x08\x12%\n\x04seed\x18\x03 \x01(\x0b\x32\x17.sarus_data_spec.Scalar\x1a\\\n\x07PushSQL\x12\x1b\n\x13\x64\x61taconnection_name\x18\x01 \x01(\t\x12\x13\n\x0bschema_name\x18\x02 \x01(\t\x12\x12\n\ntable_name\x18\x03 \x01(\t\x12\x0b\n\x03uri\x18\x04 \x01(\t\x1a\x0b\n\tSynthetic\x1a\x0e\n\x0cUserSettings\x1a/\n\x15\x41utomaticUserSettings\x12\x16\n\x0emax_categories\x18\x01 \x01(\x03\x1a\x15\n\x13PrivacyUnitTracking\x1a\x0b\n\tTranscode\x1a\x12\n\x10InverseTranscode\x1ao\n\x14\x44ifferentiatedSample\x12\x12\n\x08\x66raction\x18\x01 \x01(\x01H\x00\x12\x0e\n\x04size\x18\x02 \x01(\x03H\x00\x12%\n\x04seed\x18\x03 \x01(\x0b\x32\x17.sarus_data_spec.ScalarB\x0c\n\nproportion\x1a\x1a\n\x18PrivacyUnitTrackingPaths\x1a\r\n\x0bPublicPaths\x1a.\n\x07GetItem\x12#\n\x04path\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\x1a\x0e\n\x0c\x41ssignBudget\x1a\x11\n\x0f\x41utomaticBudget\x1a\x12\n\x10\x41ttributesBudget\x1a\n\n\x08SDBudget\x1a$\n\nDeriveSeed\x12\x16\n\x0erandom_integer\x18\x01 \x01(\x03\x1a\x0b\n\tGroupByPE\x1a\x10\n\x0eSamplingRatios\x1a\x12\n\x10RelationshipSpec\x1a\x88\x01\n\tSelectSql\x12\x0f\n\x05query\x18\x01 \x01(\tH\x00\x12\x34\n\x0f\x61liased_queries\x18\x02 \x01(\x0b\x32\x19.Transform.AliasedQueriesH\x00\x12*\n\x0bsql_dialect\x18\x03 \x01(\x0e\x32\x15.Transform.SQLDialectB\x08\n\x06select\x1a\x8a\x01\n\x0b\x44PSelectSql\x12\x0f\n\x05query\x18\x01 \x01(\tH\x00\x12\x34\n\x0f\x61liased_queries\x18\x02 \x01(\x0b\x32\x19.Transform.AliasedQueriesH\x00\x12*\n\x0bsql_dialect\x18\x03 \x01(\x0e\x32\x15.Transform.SQLDialectB\x08\n\x06select\x1a@\n\x0e\x41liasedQueries\x12.\n\raliased_query\x18\x01 \x03(\x0b\x32\x17.Transform.AliasedQuery\x1a\x42\n\x0c\x41liasedQuery\x12#\n\x04path\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\x12\r\n\x05query\x18\x02 \x01(\t\x1a\x17\n\x07\x45xtract\x12\x0c\n\x04size\x18\x01 \x01(\x05\x1a\x13\n\x11ValidatedUserType\x1a\x11\n\x0f\x45rrorEstimation\x1aV\n\x08\x46itModel\x12\x12\n\nbatch_size\x18\x01 \x01(\x05\x12\x0e\n\x06\x65pochs\x18\x02 \x01(\x05\x12&\n\ttext_kind\x18\x03 \x01(\x0b\x32\x13.Transform.TextKind\x1aN\n\x08TextKind\x12\x14\n\ntext_field\x18\x01 \x01(\tH\x00\x12\x1f\n\x04\x63hat\x18\x02 \x01(\x0b\x32\x0f.Transform.ChatH\x00\x42\x0b\n\ttext_kind\x1a\x34\n\x04\x43hat\x12\x16\n\x0equestion_field\x18\x01 \x01(\t\x12\x14\n\x0c\x61nswer_field\x18\x02 \x01(\t\x1an\n\nFitModelDP\x12\x12\n\nbatch_size\x18\x01 \x01(\x05\x12\x0e\n\x06\x65pochs\x18\x03 \x01(\x05\x12\x14\n\x0cl2_norm_clip\x18\x04 \x01(\x01\x12&\n\ttext_kind\x18\x05 \x01(\x0b\x32\x13.Transform.TextKind\x1a@\n\x11GenerateFromModel\x12\x16\n\x0emax_new_tokens\x18\x01 \x01(\x05\x12\x13\n\x0btemperature\x18\x02 \x01(\x01\x1a\x32\n\x0bSelectTable\x12#\n\x04path\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\"\x80\x01\n\nSQLDialect\x12\x08\n\x04NONE\x10\x00\x12\x0c\n\x08POSTGRES\x10\x01\x12\x0e\n\nSQL_SERVER\x10\x02\x12\n\n\x06MY_SQL\x10\x03\x12\x0b\n\x07SQLLITE\x10\x04\x12\n\n\x06ORACLE\x10\x05\x12\r\n\tBIG_QUERY\x10\x06\x12\x0c\n\x08REDSHIFT\x10\x07\x12\x08\n\x04HIVE\x10\x08\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.transform_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TRANSFORM_PROPERTIESENTRY._options = None
  _TRANSFORM_PROPERTIESENTRY._serialized_options = b'8\001'
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._options = None
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_options = b'8\001'
  _TRANSFORM._serialized_start=158
  _TRANSFORM._serialized_end=5721
  _TRANSFORM_PROPERTIESENTRY._serialized_start=338
  _TRANSFORM_PROPERTIESENTRY._serialized_end=387
  _TRANSFORM_SPEC._serialized_start=390
  _TRANSFORM_SPEC._serialized_end=2288
  _TRANSFORM_EXTERNAL._serialized_start=2291
  _TRANSFORM_EXTERNAL._serialized_end=3305
  _TRANSFORM_EXTERNAL_OPIDENTIFIER._serialized_start=2405
  _TRANSFORM_EXTERNAL_OPIDENTIFIER._serialized_end=3004
  _TRANSFORM_EXTERNAL_STD._serialized_start=3006
  _TRANSFORM_EXTERNAL_STD._serialized_end=3025
  _TRANSFORM_EXTERNAL_PANDAS._serialized_start=3027
  _TRANSFORM_EXTERNAL_PANDAS._serialized_end=3049
  _TRANSFORM_EXTERNAL_NUMPY._serialized_start=3051
  _TRANSFORM_EXTERNAL_NUMPY._serialized_end=3072
  _TRANSFORM_EXTERNAL_TENSORFLOW._serialized_start=3074
  _TRANSFORM_EXTERNAL_TENSORFLOW._serialized_end=3100
  _TRANSFORM_EXTERNAL_SKLEARN._serialized_start=3102
  _TRANSFORM_EXTERNAL_SKLEARN._serialized_end=3125
  _TRANSFORM_EXTERNAL_PANDASPROFILING._serialized_start=3127
  _TRANSFORM_EXTERNAL_PANDASPROFILING._serialized_end=3158
  _TRANSFORM_EXTERNAL_XGBOOST._serialized_start=3160
  _TRANSFORM_EXTERNAL_XGBOOST._serialized_end=3183
  _TRANSFORM_EXTERNAL_SKOPT._serialized_start=3185
  _TRANSFORM_EXTERNAL_SKOPT._serialized_end=3206
  _TRANSFORM_EXTERNAL_IMBLEARN._serialized_start=3208
  _TRANSFORM_EXTERNAL_IMBLEARN._serialized_end=3232
  _TRANSFORM_EXTERNAL_SHAP._serialized_start=3234
  _TRANSFORM_EXTERNAL_SHAP._serialized_end=3254
  _TRANSFORM_EXTERNAL_SCIPY._serialized_start=3256
  _TRANSFORM_EXTERNAL_SCIPY._serialized_end=3277
  _TRANSFORM_EXTERNAL_OPTBINNING._serialized_start=3279
  _TRANSFORM_EXTERNAL_OPTBINNING._serialized_end=3305
  _TRANSFORM_IDENTITY._serialized_start=3307
  _TRANSFORM_IDENTITY._serialized_end=3317
  _TRANSFORM_VARIABLE._serialized_start=3319
  _TRANSFORM_VARIABLE._serialized_end=3361
  _TRANSFORM_COMPOSED._serialized_start=3364
  _TRANSFORM_COMPOSED._serialized_end=3533
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_start=3480
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_end=3533
  _TRANSFORM_PROJECT._serialized_start=3535
  _TRANSFORM_PROJECT._serialized_end=3587
  _TRANSFORM_FILTER._serialized_start=3589
  _TRANSFORM_FILTER._serialized_end=3636
  _TRANSFORM_SHUFFLE._serialized_start=3638
  _TRANSFORM_SHUFFLE._serialized_end=3647
  _TRANSFORM_JOIN._serialized_start=3649
  _TRANSFORM_JOIN._serialized_end=3690
  _TRANSFORM_CAST._serialized_start=3692
  _TRANSFORM_CAST._serialized_end=3735
  _TRANSFORM_SAMPLE._serialized_start=3737
  _TRANSFORM_SAMPLE._serialized_end=3834
  _TRANSFORM_SCHEMAINFERENCE._serialized_start=3836
  _TRANSFORM_SCHEMAINFERENCE._serialized_end=3954
  _TRANSFORM_SCHEMAINFERENCE_CASTPOLICY._serialized_start=3915
  _TRANSFORM_SCHEMAINFERENCE_CASTPOLICY._serialized_end=3954
  _TRANSFORM_GROUPBY._serialized_start=3956
  _TRANSFORM_GROUPBY._serialized_end=3978
  _TRANSFORM_TOSMALLDATA._serialized_start=3980
  _TRANSFORM_TOSMALLDATA._serialized_end=4071
  _TRANSFORM_PUSHSQL._serialized_start=4073
  _TRANSFORM_PUSHSQL._serialized_end=4165
  _TRANSFORM_SYNTHETIC._serialized_start=4167
  _TRANSFORM_SYNTHETIC._serialized_end=4178
  _TRANSFORM_USERSETTINGS._serialized_start=4180
  _TRANSFORM_USERSETTINGS._serialized_end=4194
  _TRANSFORM_AUTOMATICUSERSETTINGS._serialized_start=4196
  _TRANSFORM_AUTOMATICUSERSETTINGS._serialized_end=4243
  _TRANSFORM_PRIVACYUNITTRACKING._serialized_start=4245
  _TRANSFORM_PRIVACYUNITTRACKING._serialized_end=4266
  _TRANSFORM_TRANSCODE._serialized_start=4268
  _TRANSFORM_TRANSCODE._serialized_end=4279
  _TRANSFORM_INVERSETRANSCODE._serialized_start=4281
  _TRANSFORM_INVERSETRANSCODE._serialized_end=4299
  _TRANSFORM_DIFFERENTIATEDSAMPLE._serialized_start=4301
  _TRANSFORM_DIFFERENTIATEDSAMPLE._serialized_end=4412
  _TRANSFORM_PRIVACYUNITTRACKINGPATHS._serialized_start=4414
  _TRANSFORM_PRIVACYUNITTRACKINGPATHS._serialized_end=4440
  _TRANSFORM_PUBLICPATHS._serialized_start=4442
  _TRANSFORM_PUBLICPATHS._serialized_end=4455
  _TRANSFORM_GETITEM._serialized_start=4457
  _TRANSFORM_GETITEM._serialized_end=4503
  _TRANSFORM_ASSIGNBUDGET._serialized_start=4505
  _TRANSFORM_ASSIGNBUDGET._serialized_end=4519
  _TRANSFORM_AUTOMATICBUDGET._serialized_start=4521
  _TRANSFORM_AUTOMATICBUDGET._serialized_end=4538
  _TRANSFORM_ATTRIBUTESBUDGET._serialized_start=4540
  _TRANSFORM_ATTRIBUTESBUDGET._serialized_end=4558
  _TRANSFORM_SDBUDGET._serialized_start=4560
  _TRANSFORM_SDBUDGET._serialized_end=4570
  _TRANSFORM_DERIVESEED._serialized_start=4572
  _TRANSFORM_DERIVESEED._serialized_end=4608
  _TRANSFORM_GROUPBYPE._serialized_start=4610
  _TRANSFORM_GROUPBYPE._serialized_end=4621
  _TRANSFORM_SAMPLINGRATIOS._serialized_start=4623
  _TRANSFORM_SAMPLINGRATIOS._serialized_end=4639
  _TRANSFORM_RELATIONSHIPSPEC._serialized_start=4641
  _TRANSFORM_RELATIONSHIPSPEC._serialized_end=4659
  _TRANSFORM_SELECTSQL._serialized_start=4662
  _TRANSFORM_SELECTSQL._serialized_end=4798
  _TRANSFORM_DPSELECTSQL._serialized_start=4801
  _TRANSFORM_DPSELECTSQL._serialized_end=4939
  _TRANSFORM_ALIASEDQUERIES._serialized_start=4941
  _TRANSFORM_ALIASEDQUERIES._serialized_end=5005
  _TRANSFORM_ALIASEDQUERY._serialized_start=5007
  _TRANSFORM_ALIASEDQUERY._serialized_end=5073
  _TRANSFORM_EXTRACT._serialized_start=5075
  _TRANSFORM_EXTRACT._serialized_end=5098
  _TRANSFORM_VALIDATEDUSERTYPE._serialized_start=5100
  _TRANSFORM_VALIDATEDUSERTYPE._serialized_end=5119
  _TRANSFORM_ERRORESTIMATION._serialized_start=5121
  _TRANSFORM_ERRORESTIMATION._serialized_end=5138
  _TRANSFORM_FITMODEL._serialized_start=5140
  _TRANSFORM_FITMODEL._serialized_end=5226
  _TRANSFORM_TEXTKIND._serialized_start=5228
  _TRANSFORM_TEXTKIND._serialized_end=5306
  _TRANSFORM_CHAT._serialized_start=5308
  _TRANSFORM_CHAT._serialized_end=5360
  _TRANSFORM_FITMODELDP._serialized_start=5362
  _TRANSFORM_FITMODELDP._serialized_end=5472
  _TRANSFORM_GENERATEFROMMODEL._serialized_start=5474
  _TRANSFORM_GENERATEFROMMODEL._serialized_end=5538
  _TRANSFORM_SELECTTABLE._serialized_start=5540
  _TRANSFORM_SELECTTABLE._serialized_end=5590
  _TRANSFORM_SQLDIALECT._serialized_start=5593
  _TRANSFORM_SQLDIALECT._serialized_end=5721
# @@protoc_insertion_point(module_scope)



import asyncio
import logging
import typing as t

from sarus_data_spec.manager.computations.base import BaseComputation, T
import sarus_data_spec.status as stt
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)


class LocalComputation(BaseComputation[T]):
    """Base class for computations that execute
    code locally."""

    async def result_from_stage_properties(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> T:
        raise NotImplementedError

    async def pending(self, dataspec: st.DataSpec) -> st.Status:
        """If a task is pending, it should be computed"""

        task = self.launch_task(dataspec=dataspec)
        if task is not None:
            # NB: an exception raised in an asyncio task will be reraised
            # in the awaiting code
            await task
        # In the other cases, the complete_task will be reentered with
        # the pending status, resulting in a polling process
        return await self.complete_task(dataspec)

    async def processing(self, dataspec: st.DataSpec) -> st.Status:
        """If processing, wait for the task to be ready.
        Such a case can happen if another manager has taken the computation
        of the task. After a given timeout, an error is raised.
        """

        stage = await self.wait_for_computation(
            dataspec=dataspec,
            current_stage="processing",
            timeout=self.computing_manager().computation_timeout(dataspec),
            max_delay=self.computing_manager().computation_max_delay(dataspec),
        )
        if stage.processing():
            stt.error(
                dataspec=dataspec,
                manager=dataspec.manager(),
                task=self.task_name,
                properties={
                    "message": "TimeOutError:Processing time out for task"
                    f" {self.task_name} on dataspec {dataspec}",
                    "relaunch": str(True),
                },
            )
            raise stt.DataSpecErrorStatus(
                (
                    True,
                    "TimeOutError:Processing time out for task"
                    f" {self.task_name} on dataspec {dataspec}",
                )
            )
        # if the stage is an error, it is complete_task
        # that decides what to do via the error_policy
        return await self.complete_task(dataspec)

    def launch_task(self, dataspec: st.DataSpec) -> t.Optional[t.Awaitable]:
        status = self.status(dataspec)
        if status is None:
            _, is_updated = stt.processing(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
            )
            if is_updated:
                return asyncio.create_task(
                    self.prepare(dataspec),
                    name=self.task_name + dataspec.uuid(),
                )
        else:
            stage = status.task(self.task_name)
            assert stage
            if stage.pending():
                _, is_updated = stt.processing(
                    dataspec=dataspec,
                    manager=self.computing_manager(),
                    task=self.task_name,
                )
                if is_updated:
                    return asyncio.create_task(
                        self.prepare(dataspec),
                        name=self.task_name + dataspec.uuid(),
                    )
        return None

    async def prepare(self, dataspec: st.DataSpec) -> None:
        """Prepare the computation and set the status accordingly.

        It is up to the computation to define what preparing means. It can be
        computing and caching the data as well as simply checking that the
        parents are themselves ready.
        """
        raise NotImplementedError



from __future__ import annotations

import datetime
import typing as t

from sarus_data_spec.base import Referrable
from sarus_data_spec.constants import (
    BIG_DATA_TASK,
    SAMPLE_SIZE_N_LINES,
)
from sarus_data_spec.context import global_context
from sarus_data_spec.json_serialisation import SarusJSONEncoder
from sarus_data_spec.path import straight_path
import sarus_data_spec.dataset as sd
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Transform(Referrable[sp.Transform]):
    """A python class to describe transforms"""

    def prototype(self) -> t.Type[sp.Transform]:
        """Return the type of the underlying protobuf."""
        return sp.Transform

    def name(self) -> str:
        return self._protobuf.name

    def doc(self) -> str:
        return self._protobuf.doc

    def is_composed(self) -> bool:
        """Is the transform composed."""
        return self._protobuf.spec.HasField("composed")

    def is_variable(self) -> bool:
        """Is the transform a variable."""
        return self._protobuf.spec.HasField("variable")

    def spec(self) -> str:
        return t.cast(str, self._protobuf.spec.WhichOneof("spec"))

    def is_external(self) -> bool:
        """Is the transform an external operation."""
        return self._protobuf.spec.HasField("external")

    def infer_output_type(
        self,
        *arguments: t.Union[st.DataSpec, st.Transform],
        **named_arguments: t.Union[st.DataSpec, st.Transform],
    ) -> t.Tuple[str, t.Callable[[st.DataSpec], None]]:
        """Guess if the external transform output is a Dataset or a Scalar.

        Registers schema if it is a Dataset and returns the value type.
        """
        return self.manager().infer_output_type(
            self, *arguments, **named_arguments
        )

    def transforms(self) -> t.Set[st.Transform]:
        """return all transforms (and avoid infinite recursions/loops)"""

        class Transforms(st.TransformVisitor):
            visited: t.Set[st.Transform] = set()

            def all(self, visited: st.Transform) -> None:
                self.visited.add(visited)

            def composed(
                self,
                visited: st.Transform,
                transform: st.Transform,
                *arguments: st.Transform,
                **named_arguments: st.Transform,
            ) -> None:
                self.visited.add(transform)
                if transform not in self.visited:
                    transform.accept(self)
                for arg in arguments:
                    if arg not in self.visited:
                        arg.accept(self)
                for arg in named_arguments.values():
                    if arg not in self.visited:
                        arg.accept(self)

            def other(self, visited: st.Transform) -> None:
                raise ValueError(
                    "A composed transform can only have Variables "
                    "or Composed ancestors."
                )

            def variable(
                self,
                visited: st.Transform,
                name: str,
                position: int,
            ) -> None:
                return

        visitor = Transforms()
        self.accept(visitor)
        return visitor.visited

    def variables(self) -> t.Set[st.Transform]:
        """Return all the variables from a composed transform"""
        return {
            transform
            for transform in self.transforms()
            if transform.is_variable()
        }

    def compose(
        self,
        *compose_arguments: st.Transform,
        **compose_named_arguments: st.Transform,
    ) -> st.Transform:
        class Compose(st.TransformVisitor):
            visited: t.Set[st.Transform] = set()
            result: st.Transform

            def variable(
                self,
                visited: st.Transform,
                name: str,
                position: int,
            ) -> None:
                self.result = visited
                self.result = compose_named_arguments[name]

            def composed(
                self,
                visited: st.Transform,
                transform: st.Transform,
                *arguments: st.Transform,
                **named_arguments: st.Transform,
            ) -> None:
                if visited not in self.visited:
                    self.result = composed(
                        transform,
                        *(
                            arg.compose(
                                *compose_arguments, **compose_named_arguments
                            )
                            for arg in arguments
                        ),
                        **{
                            name: arg.compose(
                                *compose_arguments, **compose_named_arguments
                            )
                            for name, arg in named_arguments.items()
                        },
                    )
                    self.visited.add(visited)
                else:
                    self.result = visited

            def other(self, visited: st.Transform) -> None:
                self.result = composed(
                    visited, *compose_arguments, **compose_named_arguments
                )

            def all(self, visited: st.Transform) -> None:
                pass

        visitor = Compose()
        self.accept(visitor)
        return visitor.result

    def apply(
        self,
        *apply_arguments: st.DataSpec,
        **apply_named_arguments: st.DataSpec,
    ) -> st.DataSpec:
        class Apply(st.TransformVisitor):
            visited: t.Dict[st.Transform, st.DataSpec] = {}
            result: st.DataSpec

            def variable(
                self,
                visited: st.Transform,
                name: str,
                position: int,
            ) -> None:
                self.result = apply_named_arguments[name]
                if self.result is None:
                    raise ValueError("Cannot substitute all variables")

            def composed(
                self,
                visited: st.Transform,
                transform: st.Transform,
                *arguments: st.Transform,
                **named_arguments: st.Transform,
            ) -> None:
                if visited not in self.visited:
                    self.result = t.cast(
                        sd.Dataset,
                        sd.transformed(
                            transform,
                            *(
                                arg.apply(
                                    *apply_arguments, **apply_named_arguments
                                )
                                for arg in arguments
                            ),
                            dataspec_type=None,
                            dataspec_name=None,
                            **{
                                name: arg.apply(
                                    *apply_arguments, **apply_named_arguments
                                )
                                for name, arg in named_arguments.items()
                            },
                        ),
                    )
                    self.visited[visited] = self.result

            def other(self, visited: st.Transform) -> None:
                self.result = sd.transformed(
                    visited,
                    *apply_arguments,
                    dataspec_type=None,
                    dataspec_name=None,
                    **apply_named_arguments,
                )

            def all(self, visited: st.Transform) -> None:
                pass

        visitor = Apply()
        self.accept(visitor)
        return visitor.result

    def abstract(
        self,
        *arguments: str,
        **named_arguments: str,
    ) -> st.Transform:
        return composed(
            self,
            *(variable(name=arg) for arg in arguments),
            **{
                name: variable(name=arg)
                for name, arg in named_arguments.items()
            },
        )

    def __call__(
        self,
        *arguments: t.Union[st.Transform, st.DataSpec, int, str],
        **named_arguments: t.Union[st.Transform, st.DataSpec, int, str],
    ) -> t.Union[st.Transform, st.DataSpec]:
        """Applies the transform to another element"""
        n_transforms = 0
        n_datasets = 0
        n_variables = 0
        for arg in arguments:
            n_transforms += int(isinstance(arg, Transform))
            n_datasets += int(isinstance(arg, st.DataSpec))
            n_variables += int(isinstance(arg, int) or isinstance(arg, str))
        for arg in named_arguments.values():
            n_transforms += int(isinstance(arg, Transform))
            n_datasets += int(isinstance(arg, st.DataSpec))
            n_variables += int(isinstance(arg, int) or isinstance(arg, str))

        total = len(arguments) + len(named_arguments)
        if total == 0:
            # If no argument is passed, we consider that we should apply
            return self.apply(
                *t.cast(t.Sequence[st.DataSpec], arguments),
                **t.cast(t.Mapping[str, st.DataSpec], named_arguments),
            )
        elif n_transforms == total:
            return self.compose(
                *t.cast(t.Sequence[Transform], arguments),
                **t.cast(t.Mapping[str, Transform], named_arguments),
            )
        elif n_variables == total:
            return self.abstract(
                *t.cast(t.Sequence[str], arguments),
                **t.cast(t.Mapping[str, str], named_arguments),
            )
        elif n_transforms + n_datasets == total:
            return self.apply(
                *t.cast(t.Sequence[st.DataSpec], arguments),
                **t.cast(t.Mapping[str, st.DataSpec], named_arguments),
            )

        return self

    def __mul__(self, argument: st.Transform) -> st.Transform:
        return self.compose(argument)

    # A Visitor acceptor
    def accept(self, visitor: st.TransformVisitor) -> None:
        visitor.all(self)
        if self.is_composed():
            visitor.composed(
                self,
                t.cast(
                    Transform,
                    self.storage().referrable(
                        self._protobuf.spec.composed.transform
                    ),
                ),
                *(
                    t.cast(Transform, self.storage().referrable(transform))
                    for transform in self._protobuf.spec.composed.arguments
                ),
                **{
                    name: t.cast(
                        Transform, self.storage().referrable(transform)
                    )
                    for name, transform in self._protobuf.spec.composed.named_arguments.items()  # noqa: E501
                },
            )
        elif self.is_variable():
            var = self._protobuf.spec.variable
            visitor.variable(self, name=var.name, position=var.position)
        else:
            visitor.other(self)

    def dot(self) -> str:
        """return a graphviz representation of the transform"""

        class Dot(st.TransformVisitor):
            visited: t.Set[st.Transform] = set()
            nodes: t.Dict[str, str] = {}
            edges: t.Set[t.Tuple[str, str]] = set()

            def variable(
                self,
                visited: st.Transform,
                name: str,
                position: int,
            ) -> None:
                self.nodes[visited.uuid()] = f"{name} ({position})"

            def composed(
                self,
                visited: st.Transform,
                transform: st.Transform,
                *arguments: st.Transform,
                **named_arguments: st.Transform,
            ) -> None:
                if visited not in self.visited:
                    transform.accept(self)
                    self.nodes[visited.uuid()] = transform.name()
                    for argument in arguments:
                        self.edges.add((argument.uuid(), visited.uuid()))
                        argument.accept(self)
                    for _, argument in named_arguments.items():
                        self.edges.add((argument.uuid(), visited.uuid()))
                        argument.accept(self)
                    self.visited.add(visited)

            def other(self, visited: st.Transform) -> None:
                raise NotImplementedError

            def all(self, visited: st.Transform) -> None:
                pass

        visitor = Dot()
        self.accept(visitor)
        result = "digraph {"
        for uuid, label in visitor.nodes.items():
            result += f'\n"{uuid}" [label="{label} ({uuid[:2]})"];'
        for u1, u2 in visitor.edges:
            result += f'\n"{u1}" -> "{u2}";'
        result += "}"
        return result

    def transform_to_apply(self) -> st.Transform:
        """Return the transform of a composed transform."""
        assert self.is_composed()
        uuid = self.protobuf().spec.composed.transform
        return t.cast(st.Transform, self.storage().referrable(uuid))

    def composed_parents(
        self,
    ) -> t.Tuple[t.List[st.Transform], t.Dict[str, st.Transform]]:
        """Return the parents of a composed transform."""
        assert self.is_composed()

        args_id = self._protobuf.spec.composed.arguments
        kwargs_id = self._protobuf.spec.composed.named_arguments

        args_parents = [
            t.cast(st.Transform, self.storage().referrable(uuid))
            for uuid in args_id
        ]
        kwargs_parents = {
            name: t.cast(st.Transform, self.storage().referrable(uuid))
            for name, uuid in kwargs_id.items()
        }
        return args_parents, kwargs_parents

    def composed_callable(self) -> t.Callable[..., t.Any]:
        """Return the composed transform's equivalent callable.

        The function takes an undefined number of named arguments.
        """
        return self.manager().composed_callable(self)


# Builders
def identity() -> Transform:
    return Transform(
        sp.Transform(
            name="Identity",
            spec=sp.Transform.Spec(identity=sp.Transform.Identity()),
            inversible=True,
            schema_preserving=True,
        )
    )


def variable(name: str, position: int = 0) -> Transform:
    return Transform(
        sp.Transform(
            name="Variable",
            spec=sp.Transform.Spec(
                variable=sp.Transform.Variable(
                    name=name,
                    position=position,
                )
            ),
            inversible=True,
            schema_preserving=True,
        )
    )


def composed(
    transform: st.Transform,
    *arguments: st.Transform,
    **named_arguments: st.Transform,
) -> st.Transform:
    if transform.is_composed():
        # We want to compose simple transforms only
        return transform.compose(*arguments, **named_arguments)
    return Transform(
        sp.Transform(
            name="Composed",
            spec=sp.Transform.Spec(
                composed=sp.Transform.Composed(
                    transform=transform.uuid(),
                    arguments=(a.uuid() for a in arguments),
                    named_arguments={
                        n: a.uuid() for n, a in named_arguments.items()
                    },
                )
            ),
        )
    )


def op_identifier_from_id(id: str) -> sp.Transform.External.OpIdentifier:
    """Build an OpIdentifier protobuf message from a string identifier.

    Args:
        identifier (str): id in the form library.name (e.g. sklearn.PD_MEAN)
    """
    parts = id.split(".")
    if len(parts) != 2:
        raise ValueError(
            f"Transform ID {id} should have the format library.name"
        )
    library, name = parts

    mapping = {
        "std": sp.Transform.External.Std,
        "sklearn": sp.Transform.External.Sklearn,
        "pandas": sp.Transform.External.Pandas,
        "pandas_profiling": sp.Transform.External.PandasProfiling,
        "numpy": sp.Transform.External.Numpy,
        "tensorflow": sp.Transform.External.Tensorflow,
        "xgboost": sp.Transform.External.XGBoost,
        "skopt": sp.Transform.External.Skopt,
        "imblearn": sp.Transform.External.Imblearn,
        "shap": sp.Transform.External.Shap,
        "scipy": sp.Transform.External.Scipy,
        "optbinning": sp.Transform.External.OptBinning,
    }

    if library not in mapping.keys():
        raise ValueError(f"Unsupported library {library}")

    MsgClass = mapping[library]
    msg = sp.Transform.External.OpIdentifier()
    getattr(msg, library).CopyFrom(MsgClass(name=name))
    return msg


def transform_id(transform: st.Transform) -> str:
    """Return the transform id."""
    spec = transform.protobuf().spec
    spec_type = str(spec.WhichOneof("spec"))
    if spec_type == "external":
        library = str(spec.external.op_identifier.WhichOneof("op"))
        op_name = getattr(spec.external.op_identifier, library).name
        return f"{library}.{op_name}"
    else:
        return spec_type


def external(
    id: str,
    py_args: t.Optional[t.Dict[int, t.Any]] = None,
    py_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
    ds_args_pos: t.Optional[t.List[int]] = None,
    ds_types: t.Optional[t.Dict[t.Union[int, str], str]] = None,
) -> Transform:
    """Create an external library transform.

    Args:
        id (str): id in the form library.name (e.g. sklearn.PD_MEAN)
        py_args (Dict[int, Any]):
            the Python objects passed as arguments to the transform.
        py_kwargs (Dict[int, Any]):
            the Python objects passed as keyword arguments to the transform.
        ds_args_pos (List[int]):
            the positions of Dataspecs passed in args.
        ds_types (Dict[int | str, str]):
            the types of the Dataspecs passed as arguments.
    """
    if py_args is None:
        py_args = {}
    if py_kwargs is None:
        py_kwargs = {}
    if ds_args_pos is None:
        ds_args_pos = []
    if ds_types is None:
        ds_types = {}
    external = sp.Transform.External(
        arguments=SarusJSONEncoder.encode_bytes([]),
        named_arguments=SarusJSONEncoder.encode_bytes(
            {
                "py_args": py_args,
                "py_kwargs": py_kwargs,
                "ds_args_pos": ds_args_pos,
                "ds_types": ds_types,
            }
        ),
        op_identifier=op_identifier_from_id(id),
    )

    return Transform(
        sp.Transform(
            name=id,
            spec=sp.Transform.Spec(
                external=external,
            ),
        )
    )


def project(projection: st.Type) -> Transform:
    return Transform(
        sp.Transform(
            name="Project",
            spec=sp.Transform.Spec(
                project=sp.Transform.Project(projection=projection.protobuf())
            ),
            inversible=False,
            schema_preserving=False,
        )
    )


def filter(filter: st.Type) -> Transform:
    return Transform(
        sp.Transform(
            name="Filter",
            spec=sp.Transform.Spec(
                filter=sp.Transform.Filter(filter=filter.protobuf())
            ),
            inversible=False,
            schema_preserving=False,
        )
    )


def get_item(path: st.Path) -> st.Transform:
    return Transform(
        sp.Transform(
            name="get_item",
            spec=sp.Transform.Spec(
                get_item=sp.Transform.GetItem(path=path.protobuf())
            ),
            inversible=False,
            schema_preserving=False,
        )
    )


def select_table(path: st.Path) -> Transform:
    return Transform(
        sp.Transform(
            name="select_table",
            spec=sp.Transform.Spec(
                select_table=sp.Transform.SelectTable(path=path.protobuf())
            ),
            inversible=False,
            schema_preserving=False,
        )
    )


def shuffle() -> Transform:
    return Transform(
        sp.Transform(
            name="Shuffle",
            spec=sp.Transform.Spec(shuffle=sp.Transform.Shuffle()),
            inversible=False,
            schema_preserving=True,
        )
    )


def join(on: st.Type) -> Transform:
    return Transform(
        sp.Transform(
            name="Join",
            spec=sp.Transform.Spec(join=sp.Transform.Join(on=on.protobuf())),
            inversible=False,
            schema_preserving=False,
        )
    )


def cast(type: st.Type) -> Transform:
    return Transform(
        sp.Transform(
            name="Cast",
            spec=sp.Transform.Spec(
                cast=sp.Transform.Cast(type=type.protobuf())
            ),
            inversible=False,
            schema_preserving=False,
        )
    )


def sample(fraction_size: t.Union[float, int], seed: st.Scalar) -> Transform:
    """Transform to sample from a dataspec
    - the dataset that needs to be protected as the first arg
    - a kwarg seed"""
    return Transform(
        sp.Transform(
            name="Sample",
            spec=sp.Transform.Spec(
                sample=sp.Transform.Sample(
                    size=fraction_size,
                    seed=seed.protobuf(),
                )
                if isinstance(fraction_size, int)
                else sp.Transform.Sample(
                    fraction=fraction_size, seed=seed.protobuf()
                )
            ),
            inversible=False,
            schema_preserving=False,
        )
    )


def user_settings() -> Transform:
    """Transform to create a dataspec from
    a protected one with a new schema. It should
    be called on:
    - the dataset that needs to be protected as the first arg
    - a kwarg user_type: scalar output of automatic_user_setttings"""
    return Transform(
        sp.Transform(
            name="User Settings",
            spec=sp.Transform.Spec(user_settings=sp.Transform.UserSettings()),
            inversible=False,
            schema_preserving=False,
        )
    )


def automatic_user_settings(max_categories: int = 200) -> Transform:
    """Transform to be called on a protected dataset
    we want to change the schema. It creates a scalar
    whose value explicits the new type of the schema"""
    return Transform(
        sp.Transform(
            name="automatic_user_settings",
            spec=sp.Transform.Spec(
                automatic_user_settings=sp.Transform.AutomaticUserSettings(
                    max_categories=max_categories
                )
            ),
            inversible=False,
            schema_preserving=False,
            properties={"creation_time": str(datetime.datetime.now())},
        )
    )


def synthetic() -> Transform:
    """Synthetic transform. This transform should be
    called on a dataset with the additional following kwargs:
    -synthetic_model: a scalar of type synthetic_model
    """
    return Transform(
        sp.Transform(
            name="Synthetic data",
            spec=sp.Transform.Spec(
                synthetic=sp.Transform.Synthetic(),
            ),
            inversible=False,
            schema_preserving=True,
        )
    )


def privacy_unit_tracking() -> Transform:
    """Transform used for PUT should be called on:
    - the dataset that needs to be PUT as the first arg
    - a kwarg privacy_unit_tracking_paths: scalar specifying the paths
     to the entities to PUT
    - a kwarg public_paths: scalar specifying the paths to
    the public tables"""
    return Transform(
        sp.Transform(
            name="Protect",
            spec=sp.Transform.Spec(
                privacy_unit_tracking=sp.Transform.PrivacyUnitTracking()
            ),
            inversible=True,
            schema_preserving=False,
        )
    )


def transcode(model_properties: t.Dict[str, t.Any]) -> st.Transform:
    raise NotImplementedError("This transform has been removed")


def inverse_transcode() -> st.Transform:
    raise NotImplementedError("This transform has been removed")


def automatic_privacy_unit_tracking_paths() -> st.Transform:
    """Transform that should be called on the dataset
    that needs to be protected, it creates a scalar whose
    value will explicit the paths to protect"""
    return Transform(
        sp.Transform(
            name="automatic_privacy_unit_tracking_paths",
            spec=sp.Transform.Spec(
                privacy_unit_tracking_paths=sp.Transform.PrivacyUnitTrackingPaths()  # noqa: E501
            ),
            properties={"creation_time": str(datetime.datetime.now())},
        )
    )


def automatic_public_paths() -> st.Transform:
    """Transform that should be called on the dataset
    that needs to be protected, it creates a scalar whose
    value will explicit the paths to public entities"""
    return Transform(
        sp.Transform(
            name="automatic_public_paths",
            spec=sp.Transform.Spec(public_paths=sp.Transform.PublicPaths()),
            properties={"creation_time": str(datetime.datetime.now())},
        )
    )


def assign_budget() -> st.Transform:
    """Transform to assign a given privacy budget to a dataset.
    It is used to specify the budget to compute the attributes
    size, bounds, marginals"""

    return Transform(
        sp.Transform(
            name="budget_assignment",
            spec=sp.Transform.Spec(assign_budget=sp.Transform.AssignBudget()),
        )
    )


def automatic_budget() -> st.Transform:
    """Transform to create a scalar specifying a budget
    automatically from the dataset it is called on.
    The rule to fix the budget is set in the corresponding
    op.
    """

    return Transform(
        sp.Transform(
            name="automatic_budget",
            spec=sp.Transform.Spec(
                automatic_budget=sp.Transform.AutomaticBudget()
            ),
        )
    )


def attributes_budget() -> st.Transform:
    """Transform to create a scalar specifying an
     epsilon,delta budget for the DP attributes of a
    dataset. It is called on a scalar specifying a
    global budget for attributes+sd."""

    return Transform(
        sp.Transform(
            name="attributes_budget",
            spec=sp.Transform.Spec(
                attribute_budget=sp.Transform.AttributesBudget()
            ),
        )
    )


def sd_budget() -> st.Transform:
    """Transform to create a scalar specifying an
     epsilon,delta budget for a synthetic dataset.
    It should be called on another scalar that specifies
    a global budget (SD+DP attributes)"""

    return Transform(
        sp.Transform(
            name="sd_budget",
            spec=sp.Transform.Spec(sd_budget=sp.Transform.SDBudget()),
        )
    )


def derive_seed(random_int: int) -> st.Transform:
    """Transform to derive a seed from a master seed"""
    return Transform(
        sp.Transform(
            name="derive_seed",
            spec=sp.Transform.Spec(
                derive_seed=sp.Transform.DeriveSeed(random_integer=random_int)
            ),
        )
    )


def group_by_pe() -> st.Transform:
    """Transform that allows to group fields
    by protected entity value. This implies that
    the dataset on which the transform is
    applied should be PUP"""
    return Transform(
        sp.Transform(
            name="group_by",
            spec=sp.Transform.Spec(group_by_pe=sp.Transform.GroupByPE()),
        )
    )


def differentiated_sample(  # type: ignore[no-untyped-def]
    fraction_size: t.Union[float, int], seed=st.Scalar
) -> Transform:
    return Transform(
        sp.Transform(
            name="DifferentiatedSample",
            spec=sp.Transform.Spec(
                differentiated_sample=sp.Transform.DifferentiatedSample(
                    size=fraction_size, seed=seed.protobuf()
                )
                if isinstance(fraction_size, int)
                else sp.Transform.DifferentiatedSample(
                    fraction=fraction_size, seed=seed.protobuf()
                )
            ),
        )
    )


def to_small_data(
    size: int, random_sampling: bool, seed: st.Scalar
) -> Transform:
    return Transform(
        sp.Transform(
            name="ToSmallData",
            spec=sp.Transform.Spec(
                to_small_data=sp.Transform.ToSmallData(
                    size=size,
                    random_sampling=random_sampling,
                    seed=seed.protobuf(),
                )
            ),
            inversible=False,
            schema_preserving=True,
        )
    )


def push_sql(
    dataconnection_name: str, schema_name: str, table_name: str, uri: str
) -> Transform:
    return Transform(
        sp.Transform(
            name="push_sql",
            spec=sp.Transform.Spec(
                push_sql=sp.Transform.PushSQL(
                    dataconnection_name=dataconnection_name,
                    schema_name=schema_name,
                    table_name=table_name,
                    uri=uri,
                )
            ),
            inversible=True,
            schema_preserving=True,
        )
    )


def handle_big_data_from_name(code_name: str) -> bool:
    """
    Check if the transform with name codename can handle bigdata in SQL.
    """
    if code_name in [
        "assign_budget",
        "differentiated_sample",
        "dp_select_sql",
        "select_sql",
        "extract",
        "sample",
        "privacy_unit_tracking",
        "user_settings",
        "automatic_user_settings",
        "ToSmallData",
    ]:
        return True
    return False


def handle_big_data(transform: st.Transform) -> bool:
    """
    Check if the transform with name codename can handle bigdata in SQL.
    """
    return handle_big_data_from_name(transform.name())


def transform_to_small_data(
    dataspec: st.DataSpec, code_name: t.Optional[str], random_sampling: bool
) -> st.DataSpec:
    """
    Transform dataspec into small data if the following transform cannot
    handle bigdata.
    The following transform name is code_name.
    """
    manager = dataspec.manager()

    if manager.is_big_data(dataspec) and (
        code_name is None or not handle_big_data_from_name(code_name)
    ):
        status = manager.status(dataspec, task_name=BIG_DATA_TASK)
        assert status
        # This could be improved by taking the threshold in bytes,
        # measuring the size of one row and
        # guessing the numbers of rows we can fetch.
        big_data_threshold_lines = int(
            status.task(BIG_DATA_TASK).properties()[SAMPLE_SIZE_N_LINES]  # type:ignore
        )
        seed = global_context().generate_seed()
        small_dataspec = t.cast(
            st.DataSpec,
            to_small_data(
                size=big_data_threshold_lines,
                random_sampling=random_sampling,
                seed=seed,
            )(dataspec),
        )
    else:
        small_dataspec = dataspec
    return small_dataspec


def select_sql(
    query: t.Union[str, t.Dict[t.Union[str, t.Tuple[str]], str]],
    dialect: t.Optional[st.SQLDialect] = None,
) -> st.Transform:
    """Transform that applies a query or a batch of aliased queries to
    a dataset.Calling .schema() or .to_arrow() on a select_sql transformed
    dataset the .sql method will be invoked and the query will be executed.
    """
    sql_dialect = (
        sp.Transform.SQLDialect.POSTGRES
        if not dialect
        else sp.Transform.SQLDialect.Value(dialect.name)
    )
    if isinstance(query, str):
        select_sql = sp.Transform.SelectSql(
            query=query,
            sql_dialect=sql_dialect,
        )
    elif len(query) == 0:
        raise ValueError(
            """Transform `SelecltSQL` must be used with
            at least one query"""
        )
    else:
        queries = {
            straight_path(
                list(
                    name
                    if isinstance(name, t.Tuple)  # type: ignore
                    else (name,)
                )
            ): qry
            for (name, qry) in query.items()
        }
        select_sql = sp.Transform.SelectSql(
            aliased_queries=sp.Transform.AliasedQueries(
                aliased_query=(
                    sp.Transform.AliasedQuery(
                        path=_path.protobuf(),
                        query=qry,
                    )
                    for (_path, qry) in queries.items()
                ),
            ),
            sql_dialect=sql_dialect,
        )

    return Transform(
        sp.Transform(
            name="select_sql",
            spec=sp.Transform.Spec(select_sql=select_sql),
            inversible=False,
            schema_preserving=False,
        )
    )


def dp_select_sql(
    query: t.Union[str, t.Dict[t.Union[str, t.Tuple[str]], str]],
    dialect: t.Optional[st.SQLDialect] = None,
) -> st.Transform:
    """DP variant of select_sql transform. It should be called with a budget
    and a seed as every dp transform.
    """
    sql_dialect = (
        sp.Transform.SQLDialect.POSTGRES
        if not dialect
        else sp.Transform.SQLDialect.Value(dialect.name)
    )
    if isinstance(query, str):
        proto = sp.Transform.DPSelectSql(
            query=query,
            sql_dialect=sql_dialect,
        )
    elif len(query) == 0:
        raise ValueError(
            """Transform `SelecltSQL` must be used with
            at least one query"""
        )
    else:
        queries = {
            straight_path(
                list(
                    name
                    if isinstance(name, t.Tuple)  # type: ignore
                    else (name,)
                )
            ): qry
            for (name, qry) in query.items()
        }
        proto = sp.Transform.DPSelectSql(
            aliased_queries=sp.Transform.AliasedQueries(
                aliased_query=(
                    sp.Transform.AliasedQuery(
                        path=_path.protobuf(),
                        query=qry,
                    )
                    for (_path, qry) in queries.items()
                ),
            ),
            sql_dialect=sql_dialect,
        )

    return Transform(
        sp.Transform(
            name="dp_select_sql",
            spec=sp.Transform.Spec(dp_select_sql=proto),
            inversible=False,
            schema_preserving=False,
        )
    )


def extract(
    size: int,
) -> st.Transform:
    """Transform that should be called on a dataset from which we want to
    extract some rows from according to the size parameter and a kwargs
    random_seed, a scalar that is a seed. For now, seed and size are
    ignored and iterating on the extract transformed dataset will be as
    iterating over the parent dataset.
    """
    return Transform(
        sp.Transform(
            name="extract",
            spec=sp.Transform.Spec(extract=sp.Transform.Extract(size=size)),
            inversible=False,
            schema_preserving=True,
        )
    )


def relationship_spec() -> st.Transform:
    """Transform that allows to redefine the primary and foreign keys
    of a dataset."""
    return Transform(
        sp.Transform(
            name="relationship_spec",
            spec=sp.Transform.Spec(
                relationship_spec=sp.Transform.RelationshipSpec()
            ),
        )
    )


def validated_user_type() -> st.Transform:
    """Transform that allows to set whether the user has validated
    the schema or if some types have to be changed"""
    return Transform(
        sp.Transform(
            name="validated_user_type",
            spec=sp.Transform.Spec(
                validated_user_type=sp.Transform.ValidatedUserType()
            ),
        )
    )


def error_estimation() -> st.Transform:
    """Transform that computes 95th percentile from true value"""
    return Transform(
        sp.Transform(
            name="error_estimation",
            spec=sp.Transform.Spec(
                error_estimation=sp.Transform.ErrorEstimation()
            ),
        )
    )


def fit_model(
    batch_size: int,
    epochs: int = 1,
    text_field: t.Optional[str] = None,
    question_field: t.Optional[str] = None,
    answer_field: t.Optional[str] = None,
) -> st.Transform:
    is_text_field = text_field is not None
    is_chat = question_field is not None and answer_field is not None

    if is_chat and is_text_field:
        raise ValueError(
            "Wrong arguments: you must either fill text_field or "
            "both answer_field and question_field"
        )
    if is_chat:
        question_field = t.cast(str, question_field)
        answer_field = t.cast(str, answer_field)
        return Transform(
            sp.Transform(
                name="fit_model",
                spec=sp.Transform.Spec(
                    fit_model=sp.Transform.FitModel(
                        batch_size=batch_size,
                        epochs=epochs,
                        text_kind=sp.Transform.TextKind(
                            chat=sp.Transform.Chat(
                                question_field=question_field,
                                answer_field=answer_field,
                            )
                        ),
                    ),
                ),
            )
        )
    if is_text_field:
        text_field = t.cast(str, text_field)
        return Transform(
            sp.Transform(
                name="fit_model",
                spec=sp.Transform.Spec(
                    fit_model=sp.Transform.FitModel(
                        batch_size=batch_size,
                        epochs=epochs,
                        text_kind=sp.Transform.TextKind(text_field=text_field),
                    ),
                ),
            )
        )

    raise ValueError(
        "Wrong arguments: you must either fill text_field "
        "or both answer_field and question_field"
    )


def fit_model_dp(
    batch_size: int,
    epochs: int = 1,
    l2_norm_clip: float = 1e-2,
    text_field: t.Optional[str] = None,
    question_field: t.Optional[str] = None,
    answer_field: t.Optional[str] = None,
) -> st.Transform:
    is_text_field = text_field is not None
    is_chat = question_field is not None and answer_field is not None

    if is_chat and is_text_field:
        raise ValueError(
            "Wrong arguments: you must either fill text_field"
            " or both answer_field and question_field"
        )
    if is_chat:
        question_field = t.cast(str, question_field)
        answer_field = t.cast(str, answer_field)
        return Transform(
            sp.Transform(
                name="fit_model_dp",
                spec=sp.Transform.Spec(
                    fit_model_dp=sp.Transform.FitModelDP(
                        batch_size=batch_size,
                        l2_norm_clip=l2_norm_clip,
                        epochs=epochs,
                        text_kind=sp.Transform.TextKind(
                            chat=sp.Transform.Chat(
                                question_field=question_field,
                                answer_field=answer_field,
                            )
                        ),
                    ),
                ),
            )
        )
    if is_text_field:
        text_field = t.cast(str, text_field)
        return Transform(
            sp.Transform(
                name="fit_model_dp",
                spec=sp.Transform.Spec(
                    fit_model_dp=sp.Transform.FitModelDP(
                        batch_size=batch_size,
                        l2_norm_clip=l2_norm_clip,
                        epochs=epochs,
                        text_kind=sp.Transform.TextKind(text_field=text_field),
                    ),
                ),
            )
        )
    raise ValueError(
        "Wrong arguments: you must either fill text_field "
        "or both answer_field and question_field"
    )


def generate_from_model(
    max_new_tokens: int = 20, temperature: float = 1.0
) -> st.Transform:
    return Transform(
        sp.Transform(
            name="generate_from_model",
            spec=sp.Transform.Spec(
                generate_from_model=sp.Transform.GenerateFromModel(
                    temperature=temperature,
                    max_new_tokens=max_new_tokens,
                )
            ),
        )
    )


if t.TYPE_CHECKING:
    test_transform: st.Transform = Transform(sp.Transform())



from typing import (
    Callable,
    Collection,
    Dict,
    Optional,
    Protocol,
    Set,
    Tuple,
    Union,
    runtime_checkable,
)

from sarus_data_spec.protobuf.typing import (
    ProtobufWithUUID,
    ProtobufWithUUIDAndDatetime,
)
from sarus_data_spec.typing import DataSpec, Referrable, Referring

# We want to store objects, be able to filter on their types and keep the last
# added in some type and relating to some object


@runtime_checkable
class Storage(Protocol):
    """Storage protocol
    A Storage can store Referrable and Referring values.
    """

    def store(self, value: Referrable[ProtobufWithUUID]) -> None:
        """Write a value to store."""
        ...

    def batch_store(
        self, values: Collection[Referrable[ProtobufWithUUID]]
    ) -> None:
        """Store a collection of referrables in the storage.

        This method does not requires the objects to be provided in the graph
        order.
        """

    def referrable(self, uuid: str) -> Optional[Referrable[ProtobufWithUUID]]:
        """Read a stored value."""
        ...

    def referring(
        self,
        referred: Union[
            Referrable[ProtobufWithUUID],
            Collection[Referrable[ProtobufWithUUID]],
        ],
        type_name: Optional[str] = None,
    ) -> Collection[Referring[ProtobufWithUUID]]:
        """List all values referring to one referred."""
        ...

    def batch_referring(
        self,
        collection_referred: Collection[
            Union[
                Referrable[ProtobufWithUUID],
                Collection[Referrable[ProtobufWithUUID]],
            ]
        ],
        type_names: Optional[Collection[str]] = None,
    ) -> Optional[Dict[str, Set[Referring[ProtobufWithUUID]]]]:
        """Returns the list of all the referring
        (for multiples type_name) of several Referrables."""
        ...

    def all_sources(self, type_name: Optional[str] = None) -> Set[DataSpec]:
        """Returns a set of all sources (with type_name) in the Storage."""
        ...

    def sources(
        self,
        referring: Referrable[ProtobufWithUUID],
        type_name: Optional[str] = None,
    ) -> Set[DataSpec]:
        """List all sources."""
        ...

    def last_referring(
        self,
        referred_uuid: Collection[str],
        type_name: str,
    ) -> Optional[Referring[ProtobufWithUUIDAndDatetime]]:
        """Last value referring to one referred.

        ``last_referring`` returns the last
        ``Referring[ProtobufWithUUIDAndDatetime]``
        object the ``type_name`` of which correspond
        to the argument ``type_name``.

        A typical use is to gather the last ``Status``
        of a ``Dataset`` and a ``Manager``.

        Note that, only time aware ``Referring``
        objects can be accessed this way as
        *last* would not make sense otherwise in the context of Data Spec where
        objects are immutable and eternal.

        Keyword arguments:

        referred:
            Either a ``Referrable`` or a collection of ``Referrable``
            referred by the object we are trying to retrieve

        type_name:
            The ``type_name`` of the Data Spec object we are trying to retrieve
        """
        ...

    def update_referring_with_datetime(
        self,
        referred_uuid: Collection[str],
        type_name: str,
        update: Callable[
            [Referring[ProtobufWithUUIDAndDatetime]],
            Tuple[Referring[ProtobufWithUUIDAndDatetime], bool],
        ],
    ) -> Tuple[Referring[ProtobufWithUUIDAndDatetime], bool]:
        """Update the last referring value of a type atomically

        Update the object ``self.last_referring(referred, type_name)``
        would be returning using the ``update`` function passed as argument.

        Note that in Sarus Data Spec, *update* means: creating an object
        with a more recent timestamp as objects are all immutable and eternal
        to simplify sync, caching and parallelism.

        Therefore everything happens as if::

            update(self.last_referring(referred, type_name))

        was inserted atomically
        (no object can be inserted with a timestamp in-between).

        Keyword arguments:

        referred:
            Either a ``Referrable`` or a collection of ``Referrable``
            referred by the object we are trying to update

        type_name:
            The ``type_name`` of the dataspec object we are trying to update

        update:
            A callable computing the new object to store
            based on the last such object.
        """
        ...

    def create_referring_with_datetime(
        self,
        value: Referring[ProtobufWithUUIDAndDatetime],
        update: Callable[
            [Referring[ProtobufWithUUIDAndDatetime]],
            Tuple[Referring[ProtobufWithUUIDAndDatetime], bool],
        ],
    ) -> Tuple[Referring[ProtobufWithUUIDAndDatetime], bool]: ...

    def type_name(
        self, type_name: str
    ) -> Collection[Referrable[ProtobufWithUUID]]:
        """List all values from a given type_name."""
        ...

    def delete(self, uuid: str) -> None:
        """Delete a stored value from the database."""
        ...

    def delete_type(self, type_name: str) -> None:
        """Delete all elements of a given type_name
        from the database and all the referrings"""
        ...


@runtime_checkable
class HasStorage(Protocol):
    """Has a storage for persistent objects."""

    def storage(self) -> Storage:
        """Return a storage (usually a singleton)."""
        ...



from typing import Any, List, Optional, Tuple, Union
import warnings

from numpy import ndarray
from pandas import DataFrame
import numpy as np
import pandas as pd

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from shap import Explanation, summary_plot
    from shap.utils import OpChain
    import shap

    explanation_abs = Explanation.abs
except ModuleNotFoundError:
    OpChain = Any
    explanation_abs = Any
    Explanation = Any
    warnings.warn("Shap not available")


class shap_plots_bar(ExternalOpImplementation):
    _transform_id = "shap.SHAP_PLOTS_BAR"
    _signature = SarusSignature(
        SarusParameter(
            name="shap_values",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="max_display",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="order",
            annotation=explanation_abs,
            default=None,
            condition=STATIC,
        ),
        SarusParameter(
            name="clustering",
            annotation=Optional[Any],
            default=None,
        ),
        SarusParameter(
            name="clustering_cutoff",
            annotation=float,
            default=0.5,
        ),
        SarusParameter(
            name="merge_cohorts",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="show_data",
            annotation=str,
            default="auto",
        ),
        SarusParameter(
            name="show",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["order"] is None:
            kwargs["order"] = shap.Explanation.abs
        return shap.plots.bar(**kwargs)


class shap_waterfall(ExternalOpImplementation):
    _transform_id = "shap.SHAP_WATERFALL"
    _signature = SarusSignature(
        SarusParameter(
            name="shap_values",
            annotation=Explanation,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="max_display",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="show",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.plots.waterfall(**kwargs)


class shap_beeswarm(ExternalOpImplementation):
    _transform_id = "shap.SHAP_BEESWARM"
    _signature = SarusSignature(
        SarusParameter(
            name="shap_values",
            annotation=Explanation,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="max_display",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="show",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="color_bar",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="plot_size",
            annotation=Union[str, float, Tuple[float, float]],
            default="auto",
        ),
        SarusParameter(
            name="order",
            annotation=Optional[OpChain],
            default=None,
        ),
        SarusParameter(
            name="clustering",
            annotation=Optional[OpChain],
            default=None,
        ),
        SarusParameter(
            name="cluster_threshold",
            annotation=Optional[float],
            default=0.5,
        ),
        SarusParameter(
            name="color",
            annotation=Optional[OpChain],
            default=None,
        ),
        SarusParameter(
            name="axis_color",
            annotation=Optional[str],
            default="#333333",
        ),
        SarusParameter(
            name="alpha",
            annotation=Optional[float],
            default=1,
        ),
        SarusParameter(
            name="show",
            annotation=Optional[bool],
            default=True,
        ),
        SarusParameter(
            name="log_scale",
            annotation=Optional[bool],
            default=False,
        ),
        SarusParameter(
            name="color_bar_label",
            annotation=Optional[str],
            default="Feature value",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["order"] is None:
            kwargs["order"] = shap.Explanation.abs.mean(0)
        return shap.plots.beeswarm(**kwargs)


class shap_heatmap(ExternalOpImplementation):
    _transform_id = "shap.SHAP_HEATMAP"
    _signature = SarusSignature(
        SarusParameter(
            name="shap_values",
            annotation=Explanation,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="instance_order",
            annotation=Union[OpChain, np.ndarray],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="feature_values",
            annotation=Union[OpChain, np.ndarray],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="feature_order",
            annotation=Optional[Union[None, OpChain, np.ndarray]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="max_display",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="show",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="plot_width",
            annotation=int,
            default=8,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        if kwargs["instance_order"] is None:
            del kwargs["instance_order"]
        if kwargs["feature_values"] is None:
            del kwargs["feature_values"]
        return shap.plots.heatmap(**kwargs)


class shap_summary_plot(ExternalOpImplementation):
    _transform_id = "shap.SHAP_SUMMARY_PLOT"
    _signature = SarusSignature(
        SarusParameter(
            name="shap_values",
            annotation=np.ndarray,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="features",
            annotation=Optional[Union[np.ndarray, pd.DataFrame, List]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="max_display",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="plot_type",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="color",
            annotation=Optional[Any],
            default=None,
        ),
        SarusParameter(
            name="axis_color",
            annotation=str,
            default="#333333",
        ),
        SarusParameter(
            name="title",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="alpha",
            annotation=float,
            default=1,
        ),
        SarusParameter(
            name="show",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="sort",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="color_bar",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="plot_size",
            annotation=Union[str, float, Tuple[float, float]],
            default="auto",
        ),
        SarusParameter(
            name="layered_violin_max_num_bins",
            annotation=int,
            default=20,
        ),
        SarusParameter(
            name="class_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="class_inds",
            annotation=Optional[List[int]],
            default=None,
        ),
        SarusParameter(
            name="color_bar_label",
            annotation=str,
            default="Feature value",
        ),
        SarusParameter(
            name="cmap",
            annotation=Any,  # Adjust accordingly
            default=None,
        ),
        SarusParameter(
            name="auto_size_plot",
            annotation=Optional[bool],
            default=None,
        ),
        SarusParameter(
            name="use_log_scale",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return summary_plot(**kwargs)


class shap_dependence_plot(ExternalOpImplementation):
    _transform_id = "shap.SHAP_DEPENDENCE_PLOT"
    _signature = SarusSignature(
        SarusParameter(
            name="ind",
            annotation=Union[int, str],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shap_values",
            annotation=ndarray,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="features",
            annotation=Union[ndarray, DataFrame],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="display_features",
            annotation=Optional[Union[ndarray, DataFrame]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="interaction_index",
            annotation=Union[str, int],
            default="auto",
        ),
        SarusParameter(
            name="color",
            annotation=str,
            default="#1E88E5",
        ),
        SarusParameter(
            name="axis_color",
            annotation=str,
            default="#333333",
        ),
        SarusParameter(
            name="cmap",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="dot_size",
            annotation=int,
            default=16,
        ),
        SarusParameter(
            name="x_jitter",
            annotation=float,
            default=0,
        ),
        SarusParameter(
            name="alpha",
            annotation=float,
            default=1,
        ),
        SarusParameter(
            name="title",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="xmin",
            annotation=Optional[Union[float, str]],
            default=None,
        ),
        SarusParameter(
            name="xmax",
            annotation=Optional[Union[float, str]],
            default=None,
        ),
        SarusParameter(
            name="ax",
            annotation=Optional[Any],  # Use proper type for matplotlib axes
            default=None,
        ),
        SarusParameter(
            name="show",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.dependence_plot(**kwargs)


class ShapForcePlot(ExternalOpImplementation):
    _transform_id = "shap.SHAP_FORCE_PLOT"
    _signature = SarusSignature(
        SarusParameter(
            name="base_value",
            annotation=float,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="shap_values",
            annotation=Optional[np.ndarray],
            condition=DATASPEC | STATIC,
            default=None,
        ),
        SarusParameter(
            name="features",
            annotation=Optional[np.ndarray],
            condition=DATASPEC | STATIC,
            default=None,
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="out_names",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="link",
            annotation=str,
            default="identity",
        ),
        SarusParameter(
            name="plot_cmap",
            annotation=str,
            default="RdBu",
        ),
        SarusParameter(
            name="matplotlib",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="show",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="figsize",
            annotation=Tuple[float, float],
            default=(20, 3),
        ),
        SarusParameter(
            name="ordering_keys",
            annotation=Optional[Any],  # Adjust this type based on your needs
            default=None,
        ),
        SarusParameter(
            name="ordering_keys_time_format",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="text_rotation",
            annotation=float,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.force_plot(**kwargs)



import logging
import typing as t

from sarus_data_spec.manager.computations.base import BaseComputation, T
import sarus_data_spec.status as stt
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)


class RemoteComputation(BaseComputation[T]):
    """Base class for computations that delegate
    execution of a task to another process"""

    async def result_from_stage_properties(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> T:
        raise NotImplementedError

    async def pending(self, dataspec: st.DataSpec) -> st.Status:
        """If the status of a task is pending, delegation has been
        already done, so the manager just waits for the task to
        be completed"""

        stage = await self.wait_for_computation(
            dataspec=dataspec,
            current_stage="pending",
            timeout=self.computing_manager().computation_timeout(dataspec),
            max_delay=self.computing_manager().computation_max_delay(dataspec),
        )

        if stage.pending():
            stt.error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": (
                        "TimeOutError: Pending time out for task"
                        f" {self.task_name} on dataspec {dataspec}"
                    ),
                    "relaunch": str(True),
                },
            )
            raise stt.DataSpecErrorStatus(
                (
                    True,
                    "Pending time out for task"
                    f" {self.task_name} on dataspec {dataspec}",
                )
            )

        return await self.complete_task(dataspec)

    async def processing(self, dataspec: st.DataSpec) -> st.Status:
        """If processing, wait for the task to be ready.
        Such a case can happen if another manager has taken the computation
        of the task. After a given timeout, an error is raised.
        """

        stage = await self.wait_for_computation(
            dataspec=dataspec,
            current_stage="processing",
            timeout=self.computing_manager().computation_timeout(dataspec),
            max_delay=self.computing_manager().computation_max_delay(dataspec),
        )
        if stage.processing() or stage.pending():
            # TODO push error status to worker ?
            stt.error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": (
                        "TimeOutError: Processing time out for task"
                        f" {self.task_name} on dataspec {dataspec}"
                    ),
                    "relaunch": str(True),
                },
            )
            raise stt.DataSpecErrorStatus(
                (
                    True,
                    "Processing time out for task"
                    f" {self.task_name} on dataspec {dataspec}",
                )
            )
        # if the stage is an error, it is complete_task
        # that decides what to do
        return await self.complete_task(dataspec)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/schema.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import type_pb2 as sarus__data__spec_dot_protobuf_dot_type__pb2
from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%sarus_data_spec/protobuf/schema.proto\x12\x0fsarus_data_spec\x1a#sarus_data_spec/protobuf/type.proto\x1a#sarus_data_spec/protobuf/path.proto\"\xd3\x03\n\x06Schema\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x61taset\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12#\n\x04type\x18\x04 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12+\n\x0cprivacy_unit\x18\x05 \x01(\x0b\x32\x15.sarus_data_spec.Path\x12;\n\nproperties\x18\x06 \x03(\x0b\x32\'.sarus_data_spec.Schema.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xd9\x01\n\nHypothesis\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x61taset\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12#\n\x04type\x18\x04 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\x46\n\nproperties\x18\x05 \x03(\x0b\x32\x32.sarus_data_spec.Schema.Hypothesis.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.schema_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SCHEMA_PROPERTIESENTRY._options = None
  _SCHEMA_PROPERTIESENTRY._serialized_options = b'8\001'
  _SCHEMA_HYPOTHESIS_PROPERTIESENTRY._options = None
  _SCHEMA_HYPOTHESIS_PROPERTIESENTRY._serialized_options = b'8\001'
  _SCHEMA._serialized_start=133
  _SCHEMA._serialized_end=600
  _SCHEMA_PROPERTIESENTRY._serialized_start=331
  _SCHEMA_PROPERTIESENTRY._serialized_end=380
  _SCHEMA_HYPOTHESIS._serialized_start=383
  _SCHEMA_HYPOTHESIS._serialized_end=600
  _SCHEMA_HYPOTHESIS_PROPERTIESENTRY._serialized_start=331
  _SCHEMA_HYPOTHESIS_PROPERTIESENTRY._serialized_end=380
# @@protoc_insertion_point(module_scope)



import typing as t

import pyarrow as pa

from sarus_data_spec.constants import DATASET_SLUGNAME
from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.schema import schema
import sarus_data_spec.typing as st


class PushSQLStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        return schema(
            self.dataset,
            schema_type=parent_schema.type(),
            privacy_unit_tracking_paths=None,
            properties=parent_schema.properties(),
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
        )


class PushSQL(StandardDatasetImplementation):
    """Computes schema and arrow
    batches for a dataspec transformed by
    an push sql transform. It is used to push the data of the parents in a
    sql database.
    """

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        return await self.parent_to_arrow(batch_size=batch_size)

    async def sql_implementation(
        self,
    ) -> t.Optional[t.Dict[t.Tuple[str, ...], str]]:
        """pass query to parents"""
        return None
